"""Desctiption of AtomSiteAniso and AtomSiteAnisoL."""
import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_1_atomic_vibrations import \
    calc_beta_by_u, vibration_constraints
from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class AtomSiteAniso(ItemN):
    """
    AtomSiteAniso class.

    Mandatory attributes:
        - label
        - type_symbol
        - fract_x
        - fract_y
        - fract_z

    Optional attributes:
        - occupancy
        - adp_type
        - u_iso_or_equiv
        - u_equiv_geom_mean
        - b_iso_or_equiv
        - multiplicity
        - wyckoff_symbol
        - cartn_x
        - cartn_y
        - cartn_z

    Internal attributes:
        - scat_length_neutron

    Internal protected attributes:
        - space_group_wyckoff
        - constr_number
    """

    ATTR_MANDATORY_NAMES = ("label", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("label", )

    ATTR_OPTIONAL_NAMES = ("u_11", "u_22", "u_33", "u_12", "u_13", "u_23",
                           "b_11", "b_22", "b_33", "b_12", "b_13", "b_23")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, float, float,
                           float, float, float, float, float, float)
    ATTR_OPTIONAL_CIF = ("U_11", "U_22", "U_33", "U_12", "U_13", "U_23",
                         "B_11", "B_22", "B_33", "B_12", "B_13", "B_23")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("scat_length_neutron", )
    ATTR_INT_PROTECTED_NAMES = ("space_group_wyckoff", "constr_number")

    # parameters considered are refined parameters
    ATTR_REF = ("u_11", "u_22", "u_33", "u_12", "u_13", "u_23",
                "b_11", "b_22", "b_33", "b_12", "b_13", "b_23")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"u_11": "{:.5f}", "u_22": "{:.5f}", "u_33": "{:.5f}",
                 "u_12": "{:.5f}", "u_13": "{:.5f}", "u_23": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_site_aniso"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomSiteAniso, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_beta(self, cell):
        """
        Calculate $beta_{ij}$ from $U_{ij}$.

        Parameters
        ----------
        cell : TYPE
            DESCRIPTION.

        Returns
        -------
        beta_11 : TYPE
            DESCRIPTION.
        beta_22 : TYPE
            DESCRIPTION.
        beta_33 : TYPE
            DESCRIPTION.
        beta_12 : TYPE
            DESCRIPTION.
        beta_13 : TYPE
            DESCRIPTION.
        beta_23 : TYPE
            DESCRIPTION.

        """
        u_i = (self.u_11, self.u_22, self.u_33,
               self.u_12, self.u_13, self.u_23)
        beta_11, beta_22, beta_33, beta_12, beta_13, beta_23 = \
            calc_beta_by_u(u_i, cell)
        return beta_11, beta_22, beta_33, beta_12, beta_13, beta_23

    def apply_space_group_constraint(self, atom_site, space_group):
        """
        Constraints of the space group.

        According to table 1 in Peterse, Palm, Acta Cryst.(1966), 20, 147
        """

        l_numb = atom_site.calc_constr_number(space_group)
        label_aniso = self.label
        label = atom_site.label
        index = label.index(label_aniso)

        self.__dict__["u_11_constraint"] = False
        self.__dict__["u_22_constraint"] = False
        self.__dict__["u_33_constraint"] = False
        self.__dict__["u_12_constraint"] = False
        self.__dict__["u_13_constraint"] = False
        self.__dict__["u_23_constraint"] = False

        numb = l_numb[index]
        
        u_i = (self.u_11, self.u_22, self.u_33, self.u_12, self.u_13,
               self.u_23)
        u_sigma_i = (self.u_11_sigma, self.u_22_sigma, self.u_33_sigma,
                     self.u_12_sigma, self.u_13_sigma, self.u_23_sigma)
        u_ref_i = (self.u_11_refinement, self.u_22_refinement,
                   self.u_33_refinement, self.u_12_refinement,
                   self.u_13_refinement, self.u_23_refinement)
        
        u_i, u_sigma_i, u_ref_i, u_constr_i = vibration_constraints(
            numb, u_i, u_sigma_i, u_ref_i)

        self.__dict__["u_11"], self.__dict__["u_22"], self.__dict__["u_33"], \
            self.__dict__["u_12"], self.__dict__["u_13"], \
            self.__dict__["u_23"] = u_i

        self.__dict__["u_11_sigma"], self.__dict__["u_22_sigma"], \
            self.__dict__["u_33_sigma"], self.__dict__["u_12_sigma"], \
            self.__dict__["u_13_sigma"], self.__dict__["u_23_sigma"] = \
            u_sigma_i

        self.__dict__["u_11_refinement"], self.__dict__["u_22_refinement"], \
            self.__dict__["u_33_refinement"], \
            self.__dict__["u_12_refinement"], \
            self.__dict__["u_13_refinement"], \
            self.__dict__["u_23_refinement"] = u_ref_i

        self.__dict__["u_11_constraint"], self.__dict__["u_22_constraint"], \
            self.__dict__["u_33_constraint"], \
            self.__dict__["u_12_constraint"], \
            self.__dict__["u_13_constraint"], \
            self.__dict__["u_23_constraint"] = u_constr_i


class AtomSiteAnisoL(LoopN):
    """Description of AtomSite in loop."""

    ITEM_CLASS = AtomSiteAniso
    ATTR_INDEX = "label"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomSiteAnisoL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def calc_beta(self, cell):
        """
        Calculate $beta_{ij}$ from $U_{ij}$.

        Output:
            numpy array of beta (11, 22, 33, 12, 13, 23)
        """
        l_beta = []
        for item in self.items:
            beta_11, beta_22, beta_33, beta_12, beta_13, beta_23 = \
                item.calc_beta(cell)
            l_beta.append((beta_11, beta_22, beta_33, beta_12, beta_13,
                           beta_23))
        return numpy.array(l_beta, dtype=float)

    def apply_space_group_constraint(self, atom_site, space_group):
        """Space group constraint."""
        for item in self.items:
            item.apply_space_group_constraint(atom_site, space_group)

# s_cont = """
#  loop_
#  _atom_site_aniso_label
#  _atom_site_aniso_U_11
#  _atom_site_aniso_U_22
#  _atom_site_aniso_U_33
#  _atom_site_aniso_U_12
#  _atom_site_aniso_U_13
#  _atom_site_aniso_U_23
#   O1   .071(1) .076(1) .0342(9) .008(1)   .0051(9) -.0030(9)
#   C2   .060(2) .072(2) .047(1)  .002(2)   .013(1)  -.009(1)
#   """

# obj = AtomSiteAnisoL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["O1"], end="\n\n")



"""Module realized some mathematical functions.

Functions
---------
    - calc_m_sigma
    - calc_scalar_product_by_vectors
    - calc_scalar_product_by_complex_vectors
    - calc_modulus_sq_by_complex_vector

"""
import numpy
import copy

from numpy.linalg import LinAlgError

def estimate_inversed_hessian_matrix(func, param_0):
    """Estimate inversed Hessian matrix."""
    n_param = len(param_0)
    np_hessian = numpy.zeros(shape=(n_param, n_param), dtype=float)
    np_first_der = numpy.zeros(shape=(n_param,), dtype=float)
    chi_sq = func(param_0)
    perc = 0.01
    
    for i_p_1, p_1 in enumerate(param_0):
        delta_p_1 = perc * numpy.abs(p_1)
        if delta_p_1 < 1e-5:
            delta_p_1 = 1e-5
        param_p = copy.deepcopy(param_0)
        param_m = copy.deepcopy(param_0)
        param_p[i_p_1] += delta_p_1
        param_m[i_p_1] -= delta_p_1
        # chi_sq_p = func(param_p)
        # chi_sq_m = func(param_m)
        # der_second = (chi_sq_p + chi_sq_m - 2.*chi_sq)/(delta_p_1**2)
        # np_hessian[i_p_1, i_p_1] = der_second

        # der_first = (chi_sq_p - chi_sq_m) / (2.*delta_p_1)
        # np_first_der[i_p_1] = der_first

        param_pp = copy.deepcopy(param_0)
        param_pm = copy.deepcopy(param_0)
        param_mp = copy.deepcopy(param_0)
        param_mm = copy.deepcopy(param_0)

        param_pp[i_p_1] += delta_p_1
        param_pm[i_p_1] += delta_p_1
        param_mp[i_p_1] -= delta_p_1
        param_mm[i_p_1] -= delta_p_1
        for i_p_2, p_2 in enumerate(param_0[:i_p_1+1]):
            delta_p_2 = perc * numpy.abs(p_2)
            if delta_p_2 < 1e-5:
                delta_p_2 = 1e-5
            param_pp[i_p_2] += delta_p_2
            param_pm[i_p_2] -= delta_p_2
            param_mp[i_p_2] += delta_p_2
            param_mm[i_p_2] -= delta_p_2
            
            chi_sq_pp = func(param_pp)
            chi_sq_pm = func(param_pm)
            chi_sq_mp = func(param_mp)
            chi_sq_mm = func(param_mm)

            der_second = (chi_sq_pp + chi_sq_mm - chi_sq_pm - chi_sq_mp) / (
                4. * delta_p_1 * delta_p_2) 
            np_hessian[i_p_1, i_p_2] = der_second
            np_hessian[i_p_2, i_p_1] = der_second

            param_pp[i_p_2] -= delta_p_2
            param_pm[i_p_2] += delta_p_2
            param_mp[i_p_2] -= delta_p_2
            param_mm[i_p_2] += delta_p_2

    try:
        np_hessian_inv = numpy.linalg.inv(np_hessian)
    except LinAlgError:
        np_hessian_inv = None
    func(param_0)
    return np_hessian_inv, np_first_der




import numpy

from .unit_cell import calc_eq_ccs_by_unit_cell_parameters
from .powder_diffraction_const_wavelength import calc_gamma_nu_by_ttheta_phi

def calc_cos_ang(cell, h_1, k_1, l_1, h_2, k_2, l_2):
    """Calculate directed cosines."""
    q_1_x, q_1_y, q_1_z = cell.calc_k_loc(h_1, k_1, l_1)
    q_2_x, q_2_y, q_2_z = cell.calc_k_loc(h_2, k_2, l_2)
    q_1_sq = q_1_x*q_1_x + q_1_y*q_1_y + q_1_z*q_1_z
    q_2_sq = q_2_x*q_2_x + q_2_y*q_2_y + q_2_z*q_2_z
    q_12 = q_1_x*q_2_x + q_1_y*q_2_y + q_1_z*q_2_z
    res = q_12/(q_1_sq*q_2_sq)**0.5
    res[res > 1.] = 1.
    return res

na = numpy.newaxis

def calc_preferred_orientation_pd(
        index_hkl, texture_g1, texture_g2, texture_axis, unit_cell_parameters,
        flag_texture_g1: bool = False, flag_texture_g2: bool = False,
        flag_texture_axis: bool = False, flag_unit_cell_parameters: bool = False):
    """Preferred orintation by Modified March-Dollas model.
    """
    preferred_orientation = None 
    eq_axis_ccs, dder_eq_axis_ccs = calc_eq_ccs_by_unit_cell_parameters(
        texture_axis, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

    eq_hkl_ccs, dder_eq_hkl_ccs = calc_eq_ccs_by_unit_cell_parameters(
        index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

    cos_alpha_ax = (eq_axis_ccs[:, 0][:, na] * eq_hkl_ccs).sum(axis=0)

    sin_alpha_ax_sq = numpy.abs(1. - numpy.square(cos_alpha_ax))
    
    cos_alpha_sq = sin_alpha_ax_sq
    hh = 1./texture_g1 + (texture_g1**2 - 1./texture_g1) * cos_alpha_sq
    
    preferred_orientation = texture_g2 + (1. - texture_g2) * numpy.power(hh, -1.5)

    dder_po = {}
    if flag_texture_g2:
        dder_po["texture_g2"] = 1. - numpy.power(hh, -1.5)
    if flag_texture_g1:
        dder_po["texture_g1"] = -1.5*(1. - texture_g2) * numpy.power(hh, -2.5) * \
            (- (1.-cos_alpha_sq)/numpy.square(texture_g1) + 2*cos_alpha_sq*texture_g1)

    return preferred_orientation, dder_po


def calc_preferred_orientation_pd2d(alpha_det,
        index_hkl, texture_g1, texture_g2, texture_axis, unit_cell_parameters,
        flag_texture_g1: bool = False, flag_texture_g2: bool = False,
        flag_texture_axis: bool = False, flag_unit_cell_parameters: bool = False):
    """Preferred orintation by Modified March-Dollas model.
    """
    preferred_orientation = None 
    eq_axis_ccs, dder_eq_axis_ccs = calc_eq_ccs_by_unit_cell_parameters(
        texture_axis, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)

    eq_hkl_ccs, dder_eq_hkl_ccs = calc_eq_ccs_by_unit_cell_parameters(
        index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)


    cos_alpha_ax = (eq_axis_ccs[:, 0][:, na] * eq_hkl_ccs).sum(axis=0)
    cos_alpha_ax[cos_alpha_ax > 1.] = 1.

    c_help = 1.-cos_alpha_ax**2
    c_help[c_help < 0.] = 0.
    sin_alpha_ax = numpy.sqrt(c_help)

    cos_alpha_det = numpy.cos(alpha_det)
    sin_alpha_det = numpy.sin(alpha_det)

    cos_alpha = cos_alpha_ax[na, na,:] * cos_alpha_det[:,:,na] + sin_alpha_ax[na, na,:] * sin_alpha_det[:,:,na]
    cos_alpha_sq = numpy.square(cos_alpha)
    hh = 1./texture_g1 + (texture_g1**2 - 1./texture_g1) * cos_alpha_sq
    
    preferred_orientation = texture_g2 + (1. - texture_g2) * numpy.power(hh, -1.5)

    dder_po = {}
    if flag_texture_g2:
        dder_po["texture_g2"] = 1. - numpy.power(hh, -1.5)
    if flag_texture_g1:
        dder_po["texture_g1"] = -1.5*(1. - texture_g2) * numpy.power(hh, -2.5) * \
            (- (1.-cos_alpha_sq)/numpy.square(texture_g1) + 2*cos_alpha_sq*texture_g1)

    return preferred_orientation, dder_po


def calc_gamma_nu_for_textured_peaks(eq_axis, eq_ccs, ttheta_hkl, texture_g1):
    c_csc = numpy.sum(eq_ccs*eq_axis, axis=0)
    inv_c_h = 1./numpy.cos(0.5 * ttheta_hkl)

    if texture_g1 <= 1:
        s_phi = inv_c_h*c_csc
        s_phi[s_phi>1] = 1.
        s_phi[s_phi<-1] = -1.
        phi_max = numpy.arcsin(s_phi)
    else:
        s_csc  = numpy.sqrt(1.-numpy.square(c_csc))
        s_csc[s_csc>1.] = 1.
        s_phi = inv_c_h*s_csc
        s_phi[s_phi>1] = 1.
        s_phi[s_phi<-1] = -1.
        phi_max = numpy.arcsin(s_phi)
    gamma_hkl, nu_hkl = calc_gamma_nu_by_ttheta_phi(
        ttheta_hkl, phi_max, flag_ttheta=False, flag_phi=False)[:2]
    return gamma_hkl, nu_hkl 


import numpy
from cryspy.A_functions_base.database import DATABASE
from cryspy.A_functions_base.charge_form_factor import get_atom_name_ion_charge_shell
from cryspy.B_parent_classes.cl_4_global import GlobalN

from cryspy.A_functions_base.structure_factor import calc_bulk_susceptibility_by_dictionary
from cryspy.A_functions_base.unit_cell import calc_volume_uc_by_unit_cell_parameters
from cryspy.E_data_classes.cl_1_crystal import Crystal
from cryspy.E_data_classes.cl_2_pd import Pd
from cryspy.E_data_classes.cl_2_pd2d import Pd2d
from cryspy.E_data_classes.cl_2_diffrn import Diffrn
from cryspy.E_data_classes.cl_2_tof import TOF


COEFF_N_A_MU_B = 6.022*0.927*1000.


def calc_unit_cell_weight(crystal: Crystal) -> float:
    """
    Calculate the weight of unit cell for crystal.

    Parameters
    ----------
    crystal : Crystal
        description of the crystals.

    Returns
    -------
    float
        unit cell in atomic weight.
    """
    d_elements = DATABASE["Elements"]
    crystal.apply_constraints()
    atom_site = crystal.atom_site
    atom_multiplicity = numpy.array(atom_site.multiplicity, dtype=int)
    atom_occupancy = numpy.array(atom_site.occupancy, dtype=float)
    l_atom_type_symbol = [get_atom_name_ion_charge_shell(
        ion_name)[0].capitalize() for ion_name in atom_site.type_symbol]
    atomic_weight = numpy.array([float(d_elements[(
        "Atomic weight", atom_type)]) for atom_type in l_atom_type_symbol],
        dtype=float)

    unit_cell_weight = numpy.sum(
        atom_multiplicity*atom_occupancy*atomic_weight)
    return unit_cell_weight


def calc_density_of_crystal(crystal: Crystal) -> float:
    """
    Calculate the density of crystal.

    Parameters
    ----------
    crystal : Crystal
        description of the crystals.

    Returns
    -------
    float
        density in g/cm^3.
    """
    unit_cell_weight = calc_unit_cell_weight(crystal)
    unit_cell_parameters = crystal.cell.get_unit_cell_parameters()
    unit_cell_volume = calc_volume_uc_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)[0]
    density = unit_cell_weight/(unit_cell_volume*0.6022)  # g/cm^3
    return density


def report_powder_experiments(rcif_object: GlobalN):
    d_elements = DATABASE["Elements"]
    ls_out = []
    rcif_dict = rcif_object.get_dictionary()
    rcif_dict_keys = rcif_dict.keys()
    l_crystal_dict_keys = [
        hh for hh in rcif_dict_keys if hh.startswith("crystal_")]
    l_pd_dict_keys = [hh for hh in rcif_dict_keys if hh.startswith("pd_")]
    l_pd2d_dict_keys = [hh for hh in rcif_dict_keys if hh.startswith("pd2d_")]

    l_crystal = [
        item for item in rcif_object.items if isinstance(item, Crystal)]
    l_crystal_name = [item.data_name.lower() for item in l_crystal]

    for item in rcif_object.items:
        f_powder = isinstance(item, (Pd, Pd2d, TOF))
        f_single = isinstance(item, (Diffrn, ))
        if f_powder or f_single:
            phase = item.phase

        if f_single:
            l_phase_items = [phase, ]
        elif f_powder:
            l_phase_items = phase.items

        if f_powder or f_single:
            field = item.setup.field
            ls_out.append(f"In experiment '{item.data_name:}' \n")

            m_chi_total = numpy.zeros((3, 3), dtype=float)
            l_s_v_sq, l_v_uc, l_m_uc = [], [], []
            l_m_chi, l_s_v_m = [], []
            l_m_chi_ion = []

            for phase_item in l_phase_items:
                phase_label = phase_item.label.lower()
                if f_single:
                    phase_scale = 1.
                else:
                    phase_scale = phase_item.scale
                ind_phase = l_crystal_name.index(phase_label)
                crystal = l_crystal[ind_phase]

                # FIXME: replace by functions defined above
                crystal.apply_constraints()
                atom_site = crystal.atom_site
                atom_multiplicity = numpy.array(
                    atom_site.multiplicity, dtype=int)
                atom_occupancy = numpy.array(atom_site.occupancy, dtype=float)
                l_atom_type_symbol = [get_atom_name_ion_charge_shell(
                    ion_name)[0].capitalize() for ion_name in atom_site.type_symbol]
                atomic_weight = numpy.array([float(d_elements[(
                    "Atomic weight", atom_type)]) for atom_type in l_atom_type_symbol], dtype=float)

                unit_cell_weight = numpy.sum(
                    atom_multiplicity*atom_occupancy*atomic_weight)
                coeff = COEFF_N_A_MU_B/unit_cell_weight
                
                crystal_dict = crystal.get_dictionary()
                crystal_dict_keys = crystal_dict.keys()
                if "atom_para_sc_chi" in crystal_dict_keys:
                    dict_in_out = {}
                    flag_use_precalculated_data = False
                    bulk_susceptibility, dder = calc_bulk_susceptibility_by_dictionary(
                        crystal_dict, dict_in_out, flag_use_precalculated_data=flag_use_precalculated_data)
                    m_chi = numpy.array([
                        [bulk_susceptibility[0], bulk_susceptibility[1],
                            bulk_susceptibility[2]],
                        [bulk_susceptibility[3], bulk_susceptibility[4],
                            bulk_susceptibility[5]],
                        [bulk_susceptibility[6], bulk_susceptibility[7], bulk_susceptibility[8]]], dtype=float)
                    atom_para_index = atom_para_index = crystal_dict["atom_para_index"]
                    n_ions = numpy.sum(
                        (atom_multiplicity*atom_occupancy)[atom_para_index])
                else:
                    m_chi = numpy.zeros((3, 3), dtype=float)
                    n_ions = 1

                v_uc, dder = calc_volume_uc_by_unit_cell_parameters(
                    crystal_dict["unit_cell_parameters"], flag_unit_cell_parameters=False)

                s_v_sq = phase_scale * v_uc * v_uc
                s_v_m = phase_scale * v_uc * unit_cell_weight
                l_m_uc.append(unit_cell_weight)
                l_v_uc.append(v_uc)
                l_s_v_sq.append(s_v_sq)
                l_s_v_m.append(s_v_m)

                l_m_chi.append(m_chi*coeff)
                l_m_chi_ion.append(m_chi/n_ions)
            sum_s_v_m = numpy.sum(l_s_v_m)
            sum_s_v_sq = numpy.sum(l_s_v_sq)
            ls_out.append(f"For polycrystalline sample at field {field:.2f}T:")
            ls_out.append(
                "-----------------------------------------------------------")
            ls_out.append(
                "Phase      Density      wF  mass M      vF       M   Moment")
            ls_out.append(
                "            g/cm^3           emu/g        emu/cm^3 mu_B/ion")
            ls_out.append(
                "-----------------------------------------------------------")

            m_chi_mixture = 0.
            m_volume_mixture = 0.
            for phase_item, v_uc, m_uc, s_v_sq, m_chi, s_v_m, m_chi_ion in zip(
                    l_phase_items, l_v_uc, l_m_uc, l_s_v_sq, l_m_chi, l_s_v_m,
                    l_m_chi_ion):
                volume_fraction = s_v_sq/sum_s_v_sq
                weight_fraction = s_v_m/sum_s_v_m
                m_chi_aver = field*(m_chi[0, 0]+m_chi[1, 1]+m_chi[2, 2])/3
                density = m_uc/(v_uc*0.6022)  # g/cm^3
                m_volume_aver = m_chi_aver*density
                m_chi_ion_aver = field * \
                    (m_chi_ion[0, 0]+m_chi_ion[1, 1]+m_chi_ion[2, 2])/3

                ls_out.append(
                    f"{phase_item.label:10} {density:7.3f}  {weight_fraction*100:5.1f}% {m_chi_aver:7.0f}  {volume_fraction*100:5.1f}% {m_volume_aver:7.0f} {m_chi_ion_aver:8.3f}")
                m_chi_mixture += weight_fraction*m_chi_aver
                m_volume_mixture += volume_fraction*m_volume_aver
            ls_out.append(
                "-----------------------------------------------------------")
            ls_out.append(
                "  wF is weight-fraction of given phase in a mixture;")
            ls_out.append("  mass M is mass magnetization of powder sample;")
            ls_out.append(
                "  vF is volume-fraction of given phase in a mixture;")
            ls_out.append("   M is volume magnetization of powder sample.\n")
            ls_out.append(
                f"Mass magnetization of a mixture is {m_chi_mixture:.2f} [emu/g];")
            ls_out.append(
                f"Volume magnetization of a mixture is {m_volume_mixture:.2f} [emu/cm^3].\n")

            ls_out.append(
                "----------------------------------------------------")
            ls_out.append(
                "Phase           Mass susceptibility tensor x 10 000")
            ls_out.append(
                "----------------------------------------------------")
            for phase_item, m_chi in zip(l_phase_items, l_m_chi):
                ls_out.append(
                    f"{phase_item.label:10} {m_chi[0, 0]:6.2f} {m_chi[1, 1]:6.2f} {m_chi[2, 2]:6.2f} {m_chi[0, 1]:6.2f} {m_chi[0, 2]:6.2f} {m_chi[1, 2]:6.2f}")
            ls_out.append(
                "----------------------------------------------------")
            ls_out.append(
                "The tensor is defined in Cartezian coordinate system")
            ls_out.append(
                "such as axis Z || c, axis X || a* for each phase in ")
            ls_out.append("[emu/(Oe g)].")
    return "\n".join(ls_out)



"""Common functions from MEM calculations.
"""

from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_1_matrices import \
    calc_product_matrix_vector, calc_vector_product, calc_mRmCmRT

#FIXME delete as it is doubled in symmetry_elements
def calc_asymmetric_unit_cell_indexes(n_x: int, n_y: int, n_z: int, r_ij, b_i)\
       -> NoReturn:
    """
    Calculate indexes of asymmetric unit cell.

    Input parameters:
        - symmetry elements;
        - points number.

    Multiplication of points number on corresponding symmetry element should
    give integer number.

    """
    r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33 = r_ij
    b_1, b_2, b_3 = b_i

    ind_x = numpy.array(range(n_x), dtype=int)
    ind_y = numpy.array(range(n_y), dtype=int)
    ind_z = numpy.array(range(n_z), dtype=int)

    ind_2d_x, ind_2d_y, ind_2d_z = numpy.meshgrid(ind_x, ind_y, ind_z,
                                                  indexing="ij")

    ind_x, ind_y = ind_2d_x.flatten(), ind_2d_y.flatten()
    ind_z = ind_2d_z.flatten()

    r_ij_2d = (
        r_11[numpy.newaxis, :], r_12[numpy.newaxis, :], r_13[numpy.newaxis, :],
        r_21[numpy.newaxis, :], r_22[numpy.newaxis, :], r_23[numpy.newaxis, :],
        r_31[numpy.newaxis, :], r_32[numpy.newaxis, :], r_33[numpy.newaxis, :])
    nb_i_2d = (numpy.around(n_x*b_1, 0).astype(int)[numpy.newaxis, :],
               numpy.around(n_y*b_2, 0).astype(int)[numpy.newaxis, :],
               numpy.around(n_z*b_3, 0).astype(int)[numpy.newaxis, :])

    ind_xyz = (ind_x[:, numpy.newaxis], ind_y[:, numpy.newaxis],
               ind_z[:, numpy.newaxis])

    ind_2d_a, ind_2d_b, ind_2d_c = calc_product_matrix_vector(r_ij_2d, ind_xyz)

    ind_2d_a += nb_i_2d[0]
    ind_2d_b += nb_i_2d[1]
    ind_2d_c += nb_i_2d[2]
    ind_2d_a = numpy.mod(ind_2d_a.astype(int), n_x)
    ind_2d_b = numpy.mod(ind_2d_b.astype(int), n_y)
    ind_2d_c = numpy.mod(ind_2d_c.astype(int), n_z)

    ind_2d_abc = n_z*n_y*ind_2d_a + n_z*ind_2d_b + ind_2d_c

    ind_2d_abc_sorted = numpy.sort(ind_2d_abc, axis=1)
    a, ind_a_u_c, counts_a_u_c = numpy.unique(
        ind_2d_abc_sorted[:, 0], return_index=True, return_counts=True)

    ind_x_a_u_c = ind_x[ind_a_u_c]
    ind_y_a_u_c = ind_y[ind_a_u_c]
    ind_z_a_u_c = ind_z[ind_a_u_c]

    return ind_x_a_u_c, ind_y_a_u_c, ind_z_a_u_c, counts_a_u_c


def calc_index_atom_symmetry_closest_to_fract_xyz(
        fract_xyz, fract_atom_xyz, r_ij, b_i, cell):
    """
    Calculate index of atoms and applied symmetry to have closest atoms.

    Basins are defined as closest points to atom

    fract_xyz = [n_points, 3]
    fract_atom_xyz = [n_atoms, 3]
    r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
    b_i = (b_1, b_2, b_3)
    cell = Cell(length_a, length_b, length_c, angle_alpha, angle_beta,
                angle_gamma)


    Output:
        - n_atom_index = [n_points]: integers from 0 until n_atoms
        - n_symmetry = [n_points]: integers from 0 until n_symmetry
    """
    r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33 = r_ij
    b_1, b_2, b_3 = b_i
    atom_x, atom_y, atom_z = fract_atom_xyz[:, 0], fract_atom_xyz[:, 1], \
        fract_atom_xyz[:, 2]
    fract_x, fract_y, fract_z = numpy.mod(fract_xyz[:, 0], 1.), \
        numpy.mod(fract_xyz[:, 1], 1.), numpy.mod(fract_xyz[:, 2], 1.)

    # 2d dimension [atoms, symmetry]
    f_at_x = numpy.mod(
        r_11[numpy.newaxis, :]*atom_x[:, numpy.newaxis] +
        r_12[numpy.newaxis, :]*atom_y[:, numpy.newaxis] +
        r_13[numpy.newaxis, :]*atom_z[:, numpy.newaxis] +
        b_1[numpy.newaxis, :], 1.)
    f_at_y = numpy.mod(
        r_21[numpy.newaxis, :]*atom_x[:, numpy.newaxis] +
        r_22[numpy.newaxis, :]*atom_y[:, numpy.newaxis] +
        r_23[numpy.newaxis, :]*atom_z[:, numpy.newaxis] +
        b_2[numpy.newaxis, :], 1.)
    f_at_z = numpy.mod(
        r_31[numpy.newaxis, :]*atom_x[:, numpy.newaxis] +
        r_32[numpy.newaxis, :]*atom_y[:, numpy.newaxis] +
        r_33[numpy.newaxis, :]*atom_z[:, numpy.newaxis] +
        b_3[numpy.newaxis, :], 1.)

    val_x = fract_x[:, numpy.newaxis, numpy.newaxis]-f_at_x[numpy.newaxis, :,
                                                            :]
    flag_x_05 = numpy.abs(val_x) > 0.5
    flag_x_sign = numpy.where(val_x >= 0.0, -1.0, 1.0)
    val_x_1 = numpy.where(flag_x_05, val_x+flag_x_sign, val_x)

    val_y = fract_y[:, numpy.newaxis, numpy.newaxis]-f_at_y[numpy.newaxis, :,
                                                            :]
    flag_y_05 = numpy.abs(val_y) > 0.5
    flag_y_sign = numpy.where(val_y >= 0.0, -1.0, 1.0)
    val_y_1 = numpy.where(flag_y_05, val_y+flag_y_sign, val_y)

    val_z = fract_z[:, numpy.newaxis, numpy.newaxis]-f_at_z[numpy.newaxis, :,
                                                            :]
    flag_z_05 = numpy.abs(val_z) > 0.5
    flag_z_sign = numpy.where(val_z >= 0.0, -1.0, 1.0)
    val_z_1 = numpy.where(flag_z_05, val_z+flag_z_sign, val_z)

    # 3d dimension [fractions, atoms, symmetry]
    p_x, p_y, p_z = cell.calc_position_by_coordinate(val_x_1, val_y_1, val_z_1)

    dist_sq_3d = numpy.square(p_x) + numpy.square(p_y) + numpy.square(p_z)
    dist_sq_2d_over_sym = numpy.min(dist_sq_3d, axis=2)
    dist_sq_2d_over_at = numpy.min(dist_sq_3d, axis=1)
    
    ind_at = numpy.argmin(dist_sq_2d_over_sym, axis=1)
    ind_sym = numpy.argmin(dist_sq_2d_over_at, axis=1)
    dist_sq = numpy.min(dist_sq_2d_over_sym, axis=1)
    distance = numpy.sqrt(dist_sq)

    return ind_at, ind_sym, distance


def calc_factor_in_front_of_density_for_fm(mult_i, np, volume,
                                           moment_2d, phase_3d, axis=2):
    r"""
    Calculate factor in front of density for magnetic structure factor.

    [hkl, points, symmetry]

    V_i = V_uc/(Ns * Np) * mult_i * moment_2d[i, s] * phase_3d[hkl, i, s]

    F_M = V_i * den_i

    V_uc is volume of unit cell
    Ns is the number of symmetry elements
    Np is the number of points in unit cell
    i is the points in assymetric unit cell
    s is the elemnt of symmetries
    mult_i is the multiplicity of i point
    den_i is the density of i point
    moment_2d[i, s] is the moment in assymetric point i at element symmetry s
        in local coordinate system
    phase_3d[hkl, i, s] is the phase for reflection hkl in the point i for
        symmetry element s


    Output data:
        - v_i_1, v_i_2, v_i_3: [hkl, points]

    """
    # number of symmetry elements
    ns = phase_3d.shape[2]

    # [ind]
    m_rho = (volume*1./float(ns*np))*mult_i

    # [ind, symm]
    m_2d_1, m_2d_2, m_2d_3 = moment_2d
    t_2d_1, t_2d_2, t_2d_3 = m_rho[:, numpy.newaxis] * m_2d_1, \
        m_rho[:, numpy.newaxis] * m_2d_2, m_rho[:, numpy.newaxis] * m_2d_3

    # [hkl, ind, symm]
    v_hkl_3d_1 = t_2d_1[numpy.newaxis, :, :] * phase_3d[:, :, :]
    v_hkl_3d_2 = t_2d_2[numpy.newaxis, :, :] * phase_3d[:, :, :]
    v_hkl_3d_3 = t_2d_3[numpy.newaxis, :, :] * phase_3d[:, :, :]
    v_hkl_2d_1 = v_hkl_3d_1.sum(axis=axis)
    v_hkl_2d_2 = v_hkl_3d_2.sum(axis=axis)
    v_hkl_2d_3 = v_hkl_3d_3.sum(axis=axis)
    return v_hkl_2d_1, v_hkl_2d_2, v_hkl_2d_3


def calc_moment_perp(k_hkl, moment_2d):
    """
    Calculate perpendicular component to k_hkl.

    k_hkl = (k_hkl_x, k_hkl_y, k_hkl_z)                   [hkl]
    moment_2d = (moment_2d_x, moment_2d_y, moment_2d_z)   [hkl, points]
    moment_perp_2d = k_hkl x moment_2d x k_hkl
    """
    k_1, k_2, k_3 = k_hkl
    k_norm = numpy.sqrt(numpy.square(k_1) + numpy.square(k_2) +
                        numpy.square(k_3))
    k_norm[k_norm == 0.] = 1.
    k_1, k_2, k_3 = k_1/k_norm, k_2/k_norm, k_3/k_norm
    # k_hkl_2d = (k_1[:, newaxis], k_2[:, newaxis], k_3[:, newaxis])
    h_1 = calc_vector_product(moment_2d, k_hkl)
    moment_perp = calc_vector_product(k_hkl, h_1)
    return moment_perp


def calc_fm_3d_by_density(mult_i, den_i, np, volume, moment_2d, phase_3d):
    """
    Calculate magnetic structure factor.

    [hkl, points, symmetry]

    F_M = V_uc / (Ns * Np) mult_i den_i moment_2d[i, s] * phase_3d[hkl, i, s]

    V_uc is volume of unit cell
    Ns is the number of symmetry elements
    Np is the number of points in unit cell
    i is the points in assymetric unit cell
    s is the elemnt of symmetries
    mult_i is the multiplicity of i point
    den_i is the density of i point
    moment_2d[i, s] is the moment in assymetric point i at element symmetry s
        in local coordinate system
    phase_3d[hkl, i, s] is the phase for reflection hkl in the point i for
        symmetry element s


    Output data:
        - f_hkl_1d_1, f_hkl_1d_2, f_hkl_1d_3: [hkl]

    """
    # number of symmetry elements
    ns = phase_3d.shape[2]

    # [ind]
    m_rho = (volume*1./float(ns*np))*den_i*mult_i

    m_2d_1, m_2d_2, m_2d_3 = moment_2d

    # [ind, symm]
    t_2d_1, t_2d_2, t_2d_3 = m_rho[:, numpy.newaxis] * \
        m_2d_1, m_rho[:, numpy.newaxis] * m_2d_2, m_rho[:, numpy.newaxis] * \
        m_2d_3

    # [hkl, ind, symm]
    f_hkl_3d_1 = t_2d_1[numpy.newaxis, :, :] * phase_3d[:, :, :]
    f_hkl_3d_2 = t_2d_2[numpy.newaxis, :, :] * phase_3d[:, :, :]
    f_hkl_3d_3 = t_2d_3[numpy.newaxis, :, :] * phase_3d[:, :, :]
    return f_hkl_3d_1, f_hkl_3d_2, f_hkl_3d_3


def calc_fm_2d_by_density(mult_i, den_i, np, volume, moment_2d, phase_3d,
                          axis: int = 2):
    r"""
    Calculate magnetic structure factor.

    [hkl, points] at axis = 2

    F_M = V_uc / (Ns * Np) mult_i den_i \sum_{s}^{Ns} moment_2d[i, s] *
        phase_3d[hkl, i, s]

    [hkl, symmetry] at axis = 1

    F_M = V_uc / (Ns * Np) \sum_{i}^{a.u.c} mult_i den_i  moment_2d[i, s] *
        phase_3d[hkl, i, s]

    """
    f_hkl_3d_1, f_hkl_3d_2, f_hkl_3d_3 = \
        calc_fm_3d_by_density(mult_i, den_i, np, volume, moment_2d, phase_3d)
    f_hkl_2d_1 = f_hkl_3d_1.sum(axis=axis)
    f_hkl_2d_2 = f_hkl_3d_2.sum(axis=axis)
    f_hkl_2d_3 = f_hkl_3d_3.sum(axis=axis)
    return f_hkl_2d_1, f_hkl_2d_2, f_hkl_2d_3


def calc_fm_by_density(mult_i, den_i, np, volume, moment_2d, phase_3d):
    r"""
    Calculate magnetic structure factor.

    [hkl]

    F_M = V_uc / (Ns * Np) \sum_{i}^{a.u.c} mult_i den_i \sum_{s}^{Ns}
        moment_2d[i, s] * phase_3d[hkl, i, s]

    """
    f_hkl_2d_1, f_hkl_2d_2, f_hkl_2d_3 = \
        calc_fm_2d_by_density(mult_i, den_i, np, volume, moment_2d, phase_3d,
                              axis=2)
    f_hkl_1d_1, f_hkl_1d_2, f_hkl_1d_3 = \
        f_hkl_2d_1.sum(axis=1), f_hkl_2d_2.sum(axis=1), f_hkl_2d_3.sum(axis=1)
    return f_hkl_1d_1, f_hkl_1d_2, f_hkl_1d_3


def transfer_to_density_3d(np_indexes, density, n_xyz, r_ij, b_i):
    """Give 3D array of density from asymetric cell.

    Arguments
    ---------
        - np_xyz is numpy array of integer numbers
        - val_1, ... are numpy array of float numbers.
    """
    (n_x, n_y, n_z) = n_xyz
    (i_x, i_y, i_z) = np_indexes
    den_3d = numpy.zeros(n_xyz)
    for r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33, b_1, b_2, b_3 \
            in zip(*r_ij, *b_i):
        np_ind_x = numpy.mod((numpy.around((
            i_x*r_11 + i_y*r_12 + i_z*r_13 + n_x*b_1).astype(float), 0)
            ).astype(int), n_x)
        np_ind_y = numpy.mod((numpy.around((
            i_x*r_21 + i_y*r_22 + i_z*r_23 + n_y*b_2).astype(float), 0)
            ).astype(int), n_y)
        np_ind_z = numpy.mod((numpy.around((
            i_x*r_31 + i_y*r_32 + i_z*r_33 + n_z*b_3).astype(float), 0)
            ).astype(int), n_z)
        den_3d[np_ind_x, np_ind_y, np_ind_z] = density
    return den_3d


def transfer_to_chi_3d(np_indexes, chi_11, chi_22, chi_33, chi_12, chi_13,
                       chi_23, n_xyz, r_ij, b_i):
    """
    Give six 3D arrays of susceptibility.

    Input arguments:
        - np_xyz is numpy array of integer numbers;
        - val_1, ... are numpy array of float numbers.
    """
    (n_x, n_y, n_z) = n_xyz
    (i_x, i_y, i_z) = np_indexes
    chi_3d_11, chi_3d_22 = numpy.zeros(n_xyz), numpy.zeros(n_xyz)
    chi_3d_33, chi_3d_12 = numpy.zeros(n_xyz), numpy.zeros(n_xyz)
    chi_3d_13, chi_3d_23 = numpy.zeros(n_xyz), numpy.zeros(n_xyz)

    for r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33, b_1, b_2, b_3 \
            in zip(*r_ij, *b_i):
        np_ind_x = numpy.mod((numpy.around((
            i_x*r_11 + i_y*r_12 + i_z*r_13 + n_x*b_1).astype(float), 0)
            ).astype(int), n_x)
        np_ind_y = numpy.mod((numpy.around((
            i_x*r_21 + i_y*r_22 + i_z*r_23 + n_y*b_2).astype(float), 0)
            ).astype(int), n_y)
        np_ind_z = numpy.mod((numpy.around((
            i_x*r_31 + i_y*r_32 + i_z*r_33 + n_z*b_3).astype(float), 0)
            ).astype(int), n_z)
        chi_out = calc_mRmCmRT(
            (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33),
            (chi_11, chi_12, chi_13, chi_12, chi_22, chi_23, chi_13, chi_23,
             chi_33))
        chi_11_rot, chi_12_rot, chi_13_rot, chi_21_rot, chi_22_rot, \
            chi_23_rot, chi_31_rot, chi_32_rot, chi_33_rot = chi_out
        chi_3d_11[np_ind_x, np_ind_y, np_ind_z] = chi_11_rot
        chi_3d_22[np_ind_x, np_ind_y, np_ind_z] = chi_22_rot
        chi_3d_33[np_ind_x, np_ind_y, np_ind_z] = chi_33_rot
        chi_3d_12[np_ind_x, np_ind_y, np_ind_z] = chi_12_rot
        chi_3d_13[np_ind_x, np_ind_y, np_ind_z] = chi_13_rot
        chi_3d_23[np_ind_x, np_ind_y, np_ind_z] = chi_23_rot
    return chi_3d_11, chi_3d_22, chi_3d_33, chi_3d_12, chi_3d_13, chi_3d_23



"""Description of Pd class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"

import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_cos_ang

from cryspy.A_functions_base.matrix_operations import calc_m1_m2_m1t
from cryspy.A_functions_base.unit_cell import calc_matrix_t

from cryspy.A_functions_base.integrated_intensity_powder_diffraction import calc_powder_iint_1d_para, calc_powder_iint_1d_ordered

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.preocedures import take_items_by_class

from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_phase import PhaseL
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs
from cryspy.C_item_loop_classes.cl_1_pd_background import PdBackgroundL
from cryspy.C_item_loop_classes.cl_1_pd_instr_reflex_asymmetry import \
    PdInstrReflexAsymmetry
from cryspy.C_item_loop_classes.cl_1_pd_instr_resolution import \
    PdInstrResolution
from cryspy.C_item_loop_classes.cl_1_pd_meas import PdMeasL
from cryspy.C_item_loop_classes.cl_1_pd_proc import PdProcL
from cryspy.C_item_loop_classes.cl_1_pd_peak import PdPeakL
from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2
from cryspy.C_item_loop_classes.cl_1_range import Range
from cryspy.C_item_loop_classes.cl_1_texture import Texture, TextureL
from cryspy.C_item_loop_classes.cl_1_exclude import ExcludeL

from cryspy.E_data_classes.cl_1_crystal import Crystal
# from cryspy.E_data_classes.cl_1_mag_crystal import MagCrystal


class Pd(DataN):
    """
    Powder diffraction experiment with polarized or unpolarized neutrons (1d).

    Data items in the DIFFRN category record details about
    powder diffraction measurements.

    Methods
    -------
        - calc_profile
        - calc_chi_sq
        - calc_iint
        - _gauss_pd
        - _lor_pd
        - calc_shape_profile
        - params_to_cif
        - data_to_cif
        - calc_to_cif
        - apply_constraints

    Attributes
    ----------
        - setup, pd_instr_resolution, phase, pd_background, pd_meas (mandatory)
        - diffrn_radiation, chi2, range, extinction, pd_instr_reflex_asymmetry
          texture, exclude, pd_proc, pd_peak, refine_ls, refln_#phase_name
          refln_susceptibility_#phase_name (optional)
    """

    CLASSES_MANDATORY = (Setup, PdInstrResolution, PhaseL, PdBackgroundL,
                         PdMeasL)
    CLASSES_OPTIONAL = (DiffrnRadiation, Chi2, Range, Extinction,
                        PdInstrReflexAsymmetry, TextureL, ExcludeL, PdProcL,
                        PdPeakL, RefineLs, ReflnL, ReflnSusceptibilityL)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "pd"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(Pd, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def apply_constraints(self):
        """Apply constraints."""
        if self.pd_meas is not None:
            self.pd_meas.apply_constraints()

    def plots(self):
        if self.is_attribute("pd_proc"):
            pd_proc = self.pd_proc
            if self.is_attribute("chi2"):
                flag_up = self.chi2.up
                flag_down = self.chi2.down
                flag_sum = self.chi2.sum
                flag_diff = self.chi2.diff
            else:
                flag_up, flag_down, flag_sum = False, False, True
                flag_diff = False

            fig_s, ax_s = pd_proc.plot_sum_diff()
            ax_s = fig_s.axes[0]
            ax_hkl = fig_s.axes[1]
            ax_s.set_title(self.data_name + " - "+ax_s.title.get_text())
            y_min_s, y_max_s = ax_hkl.get_ylim()
            y_dist_s = y_max_s-y_min_s
            y_step_s = 0.

            flag_d = False

            for item in self.items:
                if isinstance(item, PdPeakL):
                    np_tth = item.numpy_ttheta
                    ax_hkl.plot(np_tth, 0.*np_tth+y_min_s-y_step_s, "|", label=item.loop_name)
                    y_step_s += 0.05*y_dist_s
            res = []
            ax_s.legend(loc='upper right')
            res.append((fig_s, ax_s))
            return res
        elif self.is_attribute("pd_meas"):
            return self.pd_meas.plots()


    def get_dictionary(self):
        """Form dictionary. See documentation module CrysPy using Jupyter notebook.
        """
        ddict = super(Pd, self).get_dictionary()
        pd_meas, range_, exclude = None, None, None

        l_obj = take_items_by_class(self, (PdMeasL, ))
        if len(l_obj) > 0:
            pd_meas = l_obj[0]

        l_obj = take_items_by_class(self, (Range, ))
        if len(l_obj) > 0:
            range_ = l_obj[0]

        l_obj = take_items_by_class(self, (ExcludeL, ))
        if len(l_obj) > 0:
            exclude = l_obj[0]

        if pd_meas is not None:
            ttheta_deg = numpy.array(pd_meas.ttheta, dtype=float)

            ttheta_min_deg = range_.ttheta_min
            ttheta_max_deg = range_.ttheta_max
            flag_in = numpy.logical_and(
                ttheta_deg >= ttheta_min_deg,
                ttheta_deg <= ttheta_max_deg)

            ddict["ttheta"] = ttheta_deg[flag_in] * numpy.pi/180.
            if pd_meas.is_attribute("intensity_plus"):
                int_plus = numpy.array(pd_meas.intensity_plus, dtype=float)[flag_in]
                s_int_plus = numpy.array(pd_meas.intensity_plus_sigma, dtype=float)[flag_in]
                int_minus = numpy.array(pd_meas.intensity_minus, dtype=float)[flag_in]
                s_int_minus = numpy.array(pd_meas.intensity_minus_sigma, dtype=float)[flag_in]
                ddict["signal_exp_plus"] = numpy.stack([int_plus, s_int_plus], axis=0)
                ddict["signal_exp_minus"] = numpy.stack([int_minus, s_int_minus], axis=0)
            else:
                int_sum = numpy.array(pd_meas.intensity, dtype=float)[flag_in]
                s_int_sum = numpy.array(pd_meas.intensity_sigma, dtype=float)[flag_in]
                ddict["signal_exp"] = numpy.stack([int_sum, s_int_sum], axis=0)

            ttheta_in_range = ttheta_deg[flag_in]
            flag_exclude = numpy.zeros(ttheta_in_range.shape, dtype=bool)
            if exclude is not None:
                for item_e in exclude.items:
                    flag_in_1 = numpy.logical_and(
                        ttheta_in_range >= item_e.ttheta_min,
                        ttheta_in_range <= item_e.ttheta_max)
                    flag_exclude = numpy.logical_or(flag_exclude, flag_in_1)
            ddict["excluded_points"] = flag_exclude

        return ddict

    def take_parameters_from_dictionary(self, ddict_diffrn, l_parameter_name: list=None, l_sigma: list=None):
        keys = ddict_diffrn.keys()
        if l_parameter_name is not None:
            parameter_label = [hh[0] for hh in l_parameter_name]
        else:
            parameter_label = []
        if "beam_polarization" in keys:
            self.diffrn_radiation.polarization = ddict_diffrn["beam_polarization"]

        if "flipper_efficiency" in keys:
            self.diffrn_radiation.efficiency = ddict_diffrn["flipper_efficiency"]

        if "phase_scale" in keys:
            hh = ddict_diffrn["phase_scale"]
            for i_item, item in enumerate(self.phase.items):
                item.scale = float(hh[i_item])

        if "texture_g1" in keys:
            pd_texture_g1 = ddict_diffrn["texture_g1"]
            pd_texture_g2 = ddict_diffrn["texture_g2"]
            pd_texture_axis = ddict_diffrn["texture_axis"]
            for i_item, item in enumerate(self.texture.items):
                item.g_1 = float(pd_texture_g1[i_item])
                item.g_2 = float(pd_texture_g2[i_item])
                item.h_ax = float(pd_texture_axis[0, i_item])
                item.k_ax = float(pd_texture_axis[1, i_item])
                item.l_ax = float(pd_texture_axis[2, i_item])

        if "phase_ig" in keys:
            hh = ddict_diffrn["phase_ig"]
            for i_item, item in enumerate(self.phase.items):
                item.igsize = float(hh[i_item])

        if "resolution_parameters" in keys:
            hh = ddict_diffrn["resolution_parameters"]
            resolution = self.pd_instr_resolution
            resolution.u = float(hh[0])
            resolution.v = float(hh[1])
            resolution.w = float(hh[2])
            resolution.x = float(hh[3])
            resolution.y = float(hh[4])

        if "asymmetry_parameters" in keys:
            hh = ddict_diffrn["asymmetry_parameters"]
            if self.is_attribute("pd_instr_reflex_asymmetry"):
                asymmetry = self.pd_instr_reflex_asymmetry
            else:
                asymmetry = PdInstrReflexAsymmetry()
                self.items.append(asymmetry)
            asymmetry.p1 = float(hh[0])
            asymmetry.p2 = float(hh[1])
            asymmetry.p3 = float(hh[2])
            asymmetry.p4 = float(hh[3])

        if "background_intensity" in keys:
            hh = ddict_diffrn["background_intensity"]
            for i_item, item in enumerate(self.pd_background.items):
                item.intensity = float(hh[i_item])

        if "offset_ttheta" in keys:
            self.setup.offset_ttheta = float(ddict_diffrn["offset_ttheta"]) * 180./numpy.pi

        if "wavelength" in keys:
            self.setup.wavelength = float(ddict_diffrn["wavelength"])

        if len(parameter_label) > 0:
            for name, sigma in zip(l_parameter_name, l_sigma):
                if name[0] == "background_intensity":
                    self.pd_background.items[name[1][0]].intensity_sigma = float(sigma)
                if name[0] == "phase_scale":
                    self.phase.items[name[1][0]].scale_sigma = float(sigma)
                if name[0] == "texture_g1":
                    self.texture.items[name[1][0]].g_1_sigma = float(sigma)
                if name[0] == "texture_g2":
                    self.texture.items[name[1][0]].g_2_sigma = float(sigma)
                if name[0] == "phase_ig":
                    self.phase.items[name[1][0]].igsize_sigma = float(sigma)
                if name[0] == "wavelength":
                    self.setup.wavelength_sigma = float(sigma)
                if name[0] == "offset_ttheta":
                    self.setup.offset_ttheta_sigma = float(sigma) * 180./numpy.pi
                if name[0] == "beam_polarization":
                    self.diffrn_radiation.polarization_sigma = float(sigma)
                if name[0] == "flipper_efficiency":
                    self.diffrn_radiation.efficiency_sigma = float(sigma)
                if name[0] == "asymmetry_parameters":
                    asymmetry = self.pd_instr_reflex_asymmetry
                    if name[1][0] == 0:
                        asymmetry.p1_sigma = float(sigma)
                    elif name[1][0] == 1:
                        asymmetry.p2_sigma = float(sigma)
                    elif name[1][0] == 2:
                        asymmetry.p3_sigma = float(sigma)
                    elif name[1][0] == 3:
                        asymmetry.p4_sigma = float(sigma)
                if name[0] == "resolution_parameters":
                    resolution = self.pd_instr_resolution
                    if name[1][0] == 0:
                        resolution.u_sigma = float(sigma)
                    elif name[1][0] == 1:
                        resolution.v_sigma = float(sigma)
                    elif name[1][0] == 2:
                        resolution.w_sigma = float(sigma)
                    elif name[1][0] == 3:
                        resolution.x_sigma = float(sigma)
                    elif name[1][0] == 4:
                        resolution.y_sigma = float(sigma)
                if name[0] == "texture_g1":
                    self.texture.items[name[1][0]].g1_sigma = float(sigma)
                if name[0] == "texture_g2":
                    self.texture.items[name[1][0]].g2_sigma = float(sigma)
                if name[0] == "texture_axis":
                    ind_param, ind_a = name[1]
                    if ind_param == 0:
                        self.texture.items[ind_a].h_ax_sigma = float(sigma)
                    if ind_param == 1:
                        self.texture.items[ind_a].k_ax_sigma = float(sigma)
                    if ind_param == 2:
                        self.texture.items[ind_a].l_ax_sigma = float(sigma)



        if (("signal_plus" in keys) and ("signal_minus" in keys)):
            pd_proc = PdProcL()
            pd_proc.numpy_ttheta = numpy.round(ddict_diffrn["ttheta"] * 180./numpy.pi, decimals=5)
            pd_proc.numpy_ttheta_corrected = numpy.round(ddict_diffrn["ttheta_corrected"] * 180./numpy.pi, decimals=5)
            pd_proc.numpy_intensity_plus_net = numpy.round(ddict_diffrn["signal_plus"], decimals=5)
            pd_proc.numpy_intensity_minus_net = numpy.round(ddict_diffrn["signal_minus"], decimals=5)
            if "signal_exp_plus" in keys:
                pd_proc.numpy_intensity_plus = numpy.round(ddict_diffrn["signal_exp_plus"][0], decimals=5)
                pd_proc.numpy_intensity_plus_sigma = numpy.round(ddict_diffrn["signal_exp_plus"][1], decimals=5)
                pd_proc.numpy_intensity_minus = numpy.round(ddict_diffrn["signal_exp_minus"][0], decimals=5)
                pd_proc.numpy_intensity_minus_sigma = numpy.round(ddict_diffrn["signal_exp_minus"][1], decimals=5)
            else:
                pd_proc.numpy_intensity = numpy.round(ddict_diffrn["signal_exp"][0], decimals=5)
                pd_proc.numpy_intensity_sigma = numpy.round(ddict_diffrn["signal_exp"][1], decimals=5)
            pd_proc.numpy_intensity_bkg_calc = numpy.round(ddict_diffrn["signal_background"], decimals=5)
            pd_proc.numpy_excluded = ddict_diffrn["excluded_points"]
            pd_proc.numpy_to_items()
            self.pd_proc = pd_proc

        l_pd_peak = []
        l_refln = []
        for item_phase in self.phase.items:
            s_label = f"dict_in_out_{item_phase.label:}"
            if s_label in keys:
                dict_crystal = ddict_diffrn[s_label]
                dict_crystal_keys = dict_crystal.keys()
                if ("index_hkl" in dict_crystal_keys):
                    index_hkl = dict_crystal["index_hkl"]
                    pd_peak = PdPeakL(loop_name = item_phase.label)
                    pd_peak.numpy_index_h = index_hkl[0]
                    pd_peak.numpy_index_k = index_hkl[1]
                    pd_peak.numpy_index_l = index_hkl[2]

                    if "iint_plus_with_factors" in dict_crystal_keys:
                        pd_peak.numpy_intensity_plus = dict_crystal["iint_plus_with_factors"]
                    if "iint_minus_with_factors" in dict_crystal_keys:
                        pd_peak.numpy_intensity_minus = dict_crystal["iint_minus_with_factors"]
                    if "multiplicity_hkl" in dict_crystal_keys:
                        pd_peak.numpy_index_mult = dict_crystal["multiplicity_hkl"]
                    if "ttheta_hkl" in dict_crystal_keys:
                        pd_peak.numpy_ttheta = numpy.round(dict_crystal["ttheta_hkl"]* 180./numpy.pi, decimals=3)
                    pd_peak.numpy_to_items()
                    l_pd_peak.append(pd_peak)

                    flag_f_squared = False
                    if (("iint_plus" in dict_crystal_keys) and ("iint_minus" in dict_crystal_keys)):
                        f_squared_calc = dict_crystal["iint_plus"] + dict_crystal["iint_minus"]
                        f_polarized_squared_calc = dict_crystal["iint_plus"] - dict_crystal["iint_minus"]
                        flag_f_squared = True

                    if ((("f_nucl" in dict_crystal_keys) or ("f_charge" in dict_crystal_keys)) or flag_f_squared):
                        refln = ReflnL(loop_name = item_phase.label)
                        refln.numpy_index_h = index_hkl[0]
                        refln.numpy_index_k = index_hkl[1]
                        refln.numpy_index_l = index_hkl[2]

                        if "f_charge" in dict_crystal_keys:
                            refln.numpy_a_calc = dict_crystal["f_charge"].real
                            refln.numpy_b_calc = dict_crystal["f_charge"].imag
                        elif "f_nucl" in dict_crystal_keys:
                            refln.numpy_a_calc = dict_crystal["f_nucl"].real
                            refln.numpy_b_calc = dict_crystal["f_nucl"].imag

                        if flag_f_squared:
                            refln.numpy_f_squared_calc = f_squared_calc
                            refln.numpy_f_polarized_squared_calc = f_polarized_squared_calc

                        refln.numpy_to_items()
                        l_refln.append(refln)




        if len(l_pd_peak) > 0:
            self.add_items(l_pd_peak)
        if len(l_refln) > 0:
            self.add_items(l_refln)

    def define_background_points(self, step_ttheta: float = 1.):
        if self.is_attribute("pd_background"):
            pd_background = self.pd_background
        else:
            pd_background = PdBackgroundL()
            self.add_items(pd_background)
        pd_meas = self.pd_meas
        pd_background.define_points(pd_meas, step_ttheta)
        return

    def estimate_background(self):
        pd_background = self.pd_background
        intensity_refinement = numpy.copy(numpy.array(
            pd_background.intensity_refinement, dtype=bool))
        pd_background.numpy_intensity_refinement = numpy.ones_like(intensity_refinement, dtype=bool)
        pd_background.numpy_to_items()
        setup=self.setup
        pd_proc = self.pd_proc
        res = pd_proc.estimate_background(
            pd_background,
            offset_ttheta=setup.offset_ttheta)
        pd_background.numpy_intensity_refinement = intensity_refinement
        pd_background.numpy_to_items()
        return res



#!/usr/bin/env python

# Copyright (c) 2018-2023 Iurii Kibalin   
# https://github.com/ikibalin/cryspy  
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# 
# * Neither the name "CrysPy" nor the names of its contributors may
#   be used to endorse or promote products derived from this software
#   without specific prior written permission.
# 
# This software is provided by the copyright holders and contributors "as
# is" and any express or implied warranties, including, but not limited
# to, the implied warranties of merchantability and fitness for a
# particular purpose are disclaimed. In no event shall the copyright owner
# or contributors be liable for any direct, indirect, incidental, special,
# exemplary, or consequential damages (including, but not limited to,
# procurement of substitute goods or services; loss of use, data, or
# profits; or business interruption) however caused and on any theory of
# liability, whether in contract, strict liability, or tort (including
# negligence or otherwise) arising in any way out of the use of this
# software, even if advised of the possibility of such damage.

import os
import os.path
from setuptools import setup, find_packages

f_name = os.path.join(os.path.dirname(__file__), "readme.rst")
with open(f_name, 'r') as f:
    long_description = f.read()

setup(
    name='cryspy',
    version='0.7.7',
    description='PNPD data analysis',
    long_description = long_description,
    author='Iurii Kibalin',
    author_email='yurikibalin@outlook.com',
    url = 'https://github.com/ikibalin/cryspy',
    license          = 'MIT License',
    keywords         = 'Polarized Neutron Diffraction Data Analysis',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],    
    include_package_data=True,
    packages=find_packages(),  #same as name
    install_requires=[
        'numpy', 
        'scipy', 
        'pycifstar',
        'matplotlib',
    ] 
)


from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Chi2(ItemN):
    """
    Choise of the experimental data for the refinement procedure.

    Attributes
    ----------
        - sum  True
        - diff True
        - up   False
        - down False
        - asymmetry False # For flip ratios
    """

    ATTR_MANDATORY_NAMES = ()
    ATTR_MANDATORY_TYPES = ()
    ATTR_MANDATORY_CIF = ()

    ATTR_OPTIONAL_NAMES = ("sum", "diff", "up", "down", "asymmetry")
    ATTR_OPTIONAL_TYPES = (bool, bool, bool, bool, bool)
    ATTR_OPTIONAL_CIF = ("sum", "diff", "up", "down", "asymmetry")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}# 'sum': True, 'diff': True, 'up': False, 'down': False
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "chi2"

    def __init__(self, **kwargs) -> NoReturn:
        super(Chi2, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def get_dictionary(self):
        res = {}
        if self.is_attribute("sum"):
            res["flag_chi_sq_sum"] = self.sum
        if self.is_attribute("diff"):
            res["flag_chi_sq_difference"] = self.diff
        if self.is_attribute("asymmetry"):
            res["flag_asymmetry"] = self.asymmetry
        return res


class Chi2L(LoopN):
    """
    Description of chi2 in loop.

    """
    ITEM_CLASS = Chi2
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(Chi2L, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

# s_cont = """
# loop_
#   _chi2_sum
#   _chi2_diff
#   _chi2_up
#   _chi2_down
#   False True  False False
#   False False True True
# """

# """


# val_2 = Cell()
# val_2.length_a = 3.
# val_2.angle_alpha = 750

# """
# obj = Chi2L.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj.get_variable_names(), end="\n\n")




# -*- coding: utf-8 -*-
"""The MEM algorithm is described."""
from typing import List
import numpy
import scipy
import scipy.optimize
import copy

from cryspy.A_functions_base.function_1_matrices import calc_chi_sq, \
    tri_linear_interpolation, calc_product_matrix_vector
from cryspy.A_functions_base.function_2_mem import calc_moment_perp, \
    calc_fm_by_density, transfer_to_density_3d, transfer_to_chi_3d

from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs
from cryspy.C_item_loop_classes.cl_1_mem_parameters import MEMParameters

from cryspy.C_item_loop_classes.cl_2_section import Section
from cryspy.C_item_loop_classes.cl_3_density_point import DensityPointL

from cryspy.E_data_classes.cl_1_crystal import Crystal
from cryspy.E_data_classes.cl_2_diffrn import Diffrn

from cryspy.D_functions_item_loop.function_1_flip_ratio import \
    calc_fm_perp_loc, calc_e_up_loc


def choose_max_clambda(c_lambda: float, den: numpy.ndarray, der: numpy.ndarray,
                       rel_diff: float = 0.1) -> float:
    """Search optimal c_lambda.

    den_new = den * exp(-c_lambda * der)
    c_lambda: (den_new - den) < rel_diff
    """
    exp_c_lambda = numpy.exp(-c_lambda*der)
    rel_max = (numpy.abs(1.-exp_c_lambda)).max()
    arg_max = (numpy.abs(1.-exp_c_lambda)).argmax()
    if (rel_max < rel_diff) & (rel_max > 0.0001 * rel_diff):
        return c_lambda
    elif (rel_max < 0.0001 * rel_diff):
        rel_diff = 0.01 * rel_diff

    # FIXME: check it.
    if exp_c_lambda[arg_max] > 1.:
        c_lambda_new = -1*numpy.log(1+rel_diff)/der[arg_max]
    else:  # <= 1
        c_lambda_new = -1*numpy.log(1-rel_diff)/der[arg_max]
    return c_lambda_new


def maximize_entropy(crystal: Crystal, l_diffrn: List[Diffrn],
                     mem_parameters: MEMParameters,
                     c_lambda: float = 1e-7, n_iterations: int = 10000,
                     disp: bool = True, d_info: dict = None) -> DensityPointL:
    """
    Collins algroritm.

    Parameters
    ----------
    crystal : Crystal
        DESCRIPTION.
    l_diffrn : TYPE
        DESCRIPTION.
    c_lambda : float, optional
        DESCRIPTION. The default is 1e-7.
    n_iterations : int, optional
        DESCRIPTION. The default is 10000.
    chi_iso_ferro : float, optional
        DESCRIPTION. The default is 0..
    chi_iso_antiferro : float, optional
        DESCRIPTION. The default is 0..
    n_x : int, optional
        DESCRIPTION. The default is 48.
    n_y : int, optional
        DESCRIPTION. The default is 48.
    n_z : int, optional
        DESCRIPTION. The default is 48.
    prior_density: str, optional
        Choose of prior density: "core" or "uniform". The default is "uniform".
    flag_two_channel: bool, optional
        DESCRIPTION. The default is False.
    disp : bool, optional
        DESCRIPTION. The default is True.

    Returns
    -------
    DensityPointL
        Magnetization density.

    """
    flag_info = d_info is not None

    if flag_info:
        d_info_keys = d_info.keys()
        if "stop" not in d_info_keys:
            d_info["stop"] = False
        if "print" not in d_info_keys:
            d_info["print"] = ""

    chi_iso_ferro = mem_parameters.chi_ferro
    chi_iso_antiferro = mem_parameters.chi_antiferro
    points_a = mem_parameters.points_a
    points_b = mem_parameters.points_b
    points_c = mem_parameters.points_c
    prior_density = mem_parameters.prior_density
    flag_two_channel = mem_parameters.method == "2channel"
    gof_desired = mem_parameters.gof_desired

    cell = crystal.cell
    space_group = crystal.space_group
    atom_site = crystal.atom_site
    space_group_symop = space_group.full_space_group_symop
    if crystal.is_attribute("atom_site_susceptibility"):
        atom_site_susceptibility = crystal.atom_site_susceptibility
        l_magnetic_labes = atom_site_susceptibility.label
    else:
        atom_site_susceptibility = None
        l_magnetic_labes  = []
    density_point = DensityPointL()
    if prior_density == "core":
        print("The prior density is core's one.", end="\r")
        if flag_info:
            d_info["print"] = "The prior density is core's one."

        atom_electron_configuration = crystal.atom_electron_configuration
        density_point.create_core_density(
            space_group_symop, cell, atom_site, atom_electron_configuration,
            points_a=points_a, points_b=points_b, points_c=points_c,
            flag_two_channel=flag_two_channel)
    else:
        print("The prior density is uniform.", end="\r")
        if flag_info:
            d_info["print"] = "The prior density is uniform."
        density_point.create_flat_density(
            space_group_symop, cell, atom_site,
            l_magnetic_labes=l_magnetic_labes, points_a=points_a,
            points_b=points_b, points_c=points_c,
            flag_two_channel=flag_two_channel)

    l_f_nucl, l_v_2d_i, l_fr_e, l_fr_s, l_fm_orb_perp_loc = [], [], [], [], []
    total_peaks = 0

    for diffrn in l_diffrn:
        diffrn_orient_matrix = diffrn.diffrn_orient_matrix
        u_matrix = diffrn_orient_matrix.u
        e_up = calc_e_up_loc(0., 0., 0., u_matrix)

        setup = diffrn.setup
        field = float(setup.field)
        h_loc = (field*e_up[0], field*e_up[1], field*e_up[2])
        diffrn_refln = diffrn.diffrn_refln
        ind_h = numpy.array(diffrn_refln.index_h, dtype=int)
        ind_k = numpy.array(diffrn_refln.index_k, dtype=int)
        ind_l = numpy.array(diffrn_refln.index_l, dtype=int)
        total_peaks += ind_h.size

        chi_m = crystal.calc_susceptibility_moment_tensor(
            ind_h, ind_k, ind_l, flag_only_orbital=True)
        sft_ij = chi_m[:9]
        sftm_ij = chi_m[9:]

        k_loc_i = cell.calc_k_loc(ind_h, ind_k, ind_l)
        fm_orb_perp_loc = calc_fm_perp_loc(e_up, field, k_loc_i, sft_ij,
                                           sftm_ij)

        hkl = (ind_h, ind_k, ind_l)
        index_hkl = numpy.stack([ind_h, ind_k, ind_l], dtype=bool)
        fr_e = numpy.array(diffrn_refln.fr, dtype=float)
        fr_s = numpy.array(diffrn_refln.fr_sigma, dtype=float)
        v_hkl_perp_2d_i, v_b_ferro, v_b_antiferro = \
            density_point.calc_factor_in_front_of_density_for_fm_perp(
                hkl, space_group_symop, cell, atom_site_susceptibility, h_loc,
                chi_iso_ferro=chi_iso_ferro,
                chi_iso_antiferro=chi_iso_antiferro,
                flag_two_channel=flag_two_channel)
        f_nucl = crystal.calc_f_nucl(index_hkl)
        l_f_nucl.append(f_nucl)
        l_v_2d_i.append((v_hkl_perp_2d_i, v_b_ferro, v_b_antiferro))
        l_fr_e.append(fr_e)
        l_fr_s.append(fr_s)
        l_fm_orb_perp_loc.append(fm_orb_perp_loc)

    def temp_func(numpy_den=None):
        l_chi_sq, l_der_chi_sq = [], []
        l_der_chi_sq_f, l_der_chi_sq_a = [], []
        for diffrn, f_nucl, v_2d_i, fr_e, fr_s, fm_orb_perp_loc in \
                zip(l_diffrn, l_f_nucl, l_v_2d_i, l_fr_e, l_fr_s,
                    l_fm_orb_perp_loc):
            diffrn_refln = diffrn.diffrn_refln

            f_m_perp, delta_f_m_perp, delta_f_m_perp_f, delta_f_m_perp_a = \
                density_point.calc_fm(*v_2d_i)

            # FIXME: put condition
            f_m_perp = (f_m_perp[0] + fm_orb_perp_loc[0],
                        f_m_perp[1] + fm_orb_perp_loc[1],
                        f_m_perp[2] + fm_orb_perp_loc[2])

            fr_m, delta_fr_m = diffrn.calc_fr(cell, f_nucl, f_m_perp,
                                              delta_f_nucl=None,
                                              delta_f_m_perp=delta_f_m_perp)
            delta_fr_m_f = diffrn.calc_fr(cell, f_nucl, f_m_perp,
                                          delta_f_nucl=None,
                                          delta_f_m_perp=delta_f_m_perp_f)[1]
            delta_fr_m_a = diffrn.calc_fr(cell, f_nucl, f_m_perp,
                                          delta_f_nucl=None,
                                          delta_f_m_perp=delta_f_m_perp_a)[1]

            diffrn_refln.numpy_fr_calc = fr_m
            chi_sq, der_chi_sq = calc_chi_sq(fr_e, fr_s, fr_m, delta_fr_m)
            der_chi_sq_f = calc_chi_sq(fr_e, fr_s, fr_m, delta_fr_m_f)[1]
            der_chi_sq_a = calc_chi_sq(fr_e, fr_s, fr_m, delta_fr_m_a)[1]
            l_chi_sq.append(chi_sq)
            l_der_chi_sq.append(der_chi_sq)
            l_der_chi_sq_f.append(der_chi_sq_f)
            l_der_chi_sq_a.append(der_chi_sq_a)
        # print(" ".join([f" {val:10.2f}" for val in l_chi_sq]))
        return sum(l_chi_sq), sum(l_der_chi_sq), sum(l_der_chi_sq_f), \
            sum(l_der_chi_sq_a)

    chi_sq_best, chi_sq_n_diff = numpy.inf, numpy.inf
    numpy_density_best, numpy_density_ferro_best = None, None
    numpy_density_antiferro_best = None
    delta_chi_sq_best, delta_chi_sq_f_best = None, None
    delta_chi_sq_a_best = None

    mult_i = density_point.numpy_multiplicity

    c_lambda_min = 1e-9  # min value
    i_cycle = 0
    while i_cycle <= n_iterations:
        i_cycle += 1
        numpy_density = density_point.numpy_density
        numpy_density_ferro = density_point.numpy_density_ferro
        numpy_density_antiferro = density_point.numpy_density_antiferro
        chi_sq, delta_chi_sq, delta_chi_sq_f, delta_chi_sq_a = temp_func()

        chi_sq_n = chi_sq/float(total_peaks)
        if disp:
            print(f"cycle {i_cycle:5}. chi_sq/n: {chi_sq_n:.2f} \
{c_lambda*10**5:.3f} {chi_sq_n_diff:.4f}",
                  end="\r")
        if flag_info:
            d_info["print"] = f"""Iteration {i_cycle:5}:
    chi_sq/n: {chi_sq_n:.2f};
    c_lambda: {c_lambda*10**5:.3f} * 10**-5;
    difference of chi_sq/n: {chi_sq_n_diff:.4f}."""

        if chi_sq > chi_sq_best:
            numpy_density = numpy_density_best
            numpy_density_ferro = numpy_density_ferro_best
            numpy_density_antiferro = numpy_density_antiferro_best
            delta_chi_sq = delta_chi_sq_best
            delta_chi_sq_f = delta_chi_sq_f_best
            delta_chi_sq_a = delta_chi_sq_a_best
            c_lambda = 0.5*c_lambda
        else:
            chi_sq_n_diff = (chi_sq_best - chi_sq)/float(total_peaks)
            chi_sq_best = chi_sq
            c_lambda = 1.03*c_lambda

        if not(flag_two_channel):
            numpy_density_best = copy.deepcopy(numpy_density)
            delta_chi_sq_best = copy.deepcopy(delta_chi_sq)
        numpy_density_ferro_best = copy.deepcopy(numpy_density_ferro)
        numpy_density_antiferro_best = copy.deepcopy(numpy_density_antiferro)
        delta_chi_sq_f_best = copy.deepcopy(delta_chi_sq_f)
        delta_chi_sq_a_best = copy.deepcopy(delta_chi_sq_a)

        if chi_sq_n < gof_desired:
            print(f"OUT: cycle {i_cycle:5} chi_sq/n is less than \
{gof_desired:.2f}",
                  end="\r")
            if flag_info:
                d_info["print"] = f"""OUT on iteration {i_cycle:5}:
    chi_sq/n is less than {gof_desired:.2f}."""
            break
        elif ((c_lambda < c_lambda_min) & False):
            print(f"OUT: cycle {i_cycle:5} chi_sq/n is {chi_sq_n:.2f}. \
c_lambda: {c_lambda:}",
                  end="\r")
            if flag_info:
                d_info["print"] = f"""OUT on iteration {i_cycle:5}:
    chi_sq/n is {chi_sq_n:.2f}. c_lambda less minimal"""
            if not(flag_two_channel):
                density_point.numpy_density = numpy_density_best
            density_point.numpy_density_ferro = numpy_density_ferro_best
            density_point.numpy_density_antiferro = \
                numpy_density_antiferro_best
            density_point.renormalize_numpy_densities(
                flag_two_channel=flag_two_channel)
            break
        elif ((chi_sq_n_diff < 0.001) and (i_cycle > 10)):
            print(f"OUT: cycle {i_cycle:5} chi_sq/n is {chi_sq_n:.2f} as diff \
of GoF is less than 0.001.",
                  end="\r")
            if flag_info:
                d_info["print"] = f"""OUT on iteration {i_cycle:5}:
    chi_sq/n is {chi_sq_n:.2f};
    difference of GoF is less than 0.001."""
            break
        elif flag_info:
            if d_info["stop"]:
                d_info["stop"] = False
                break

        delta_chi_sq_f_mult_i = delta_chi_sq_f/mult_i
        delta_chi_sq_a_mult_i = delta_chi_sq_a/mult_i

        rel_diff = 0.05
        if not(flag_two_channel):
            delta_chi_sq_mult_i = delta_chi_sq/mult_i
            c_lambda = choose_max_clambda(c_lambda, numpy_density,
                                          delta_chi_sq_mult_i, rel_diff)
            density_point.numpy_density = numpy_density*numpy.exp(
                -c_lambda*delta_chi_sq_mult_i)
        else:
            c_lambda = choose_max_clambda(c_lambda, numpy_density_ferro,
                                          delta_chi_sq_f_mult_i, rel_diff)

        density_point.numpy_density_ferro = \
            numpy_density_ferro*numpy.exp(-c_lambda*delta_chi_sq_f_mult_i)
        density_point.numpy_density_antiferro = \
            numpy_density_antiferro*numpy.exp(-c_lambda*delta_chi_sq_a_mult_i)

        density_point.renormalize_numpy_densities(
            flag_two_channel=flag_two_channel)

    density_point.numpy_to_items()
    for diffrn in l_diffrn:
        diffrn.diffrn_refln.numpy_to_items()
        chi_sq, points = diffrn.diffrn_refln.calc_chi_sq_points()
        refine_ls = RefineLs(goodness_of_fit_all=chi_sq/points,
                             number_reflns=points)
        diffrn.add_items([refine_ls])
    return density_point


def refine_susceptibility(crystal: Crystal, l_diffrn: List[Diffrn],
                          density_point: DensityPointL,
                          mem_parameters: MEMParameters, disp: bool = True,
                          d_info: dict = None) -> (float, float):
    """
    Refinement of susceptibility.

    Parameters
    ----------
    crystal : Crystal
        DESCRIPTION.
    l_diffrn : TYPE
        DESCRIPTION.
    density_point : DensityPointL
        DESCRIPTION.
    chi_iso_ferro : float, optional
        DESCRIPTION. The default is 0..
    chi_iso_antiferro : float, optional
        DESCRIPTION. The default is 0..
    flag_ferro : bool, optional
        DESCRIPTION. The default is True.
    flag_antiferro : bool, optional
        DESCRIPTION. The default is True.
    disp : bool, optional
        DESCRIPTION. The default is True.

    Returns
    -------
    float
        Chi_iso_ferro.
    float
        Chi_iso_antiferro.

    """
    flag_info = d_info is not None

    if flag_info:
        d_info_keys = d_info.keys()
        if "stop" not in d_info_keys:
            d_info["stop"] = False
        if "print" not in d_info_keys:
            d_info["print"] = ""

    crystal.apply_constraints()

    flag_two_channel = mem_parameters.method == "2channel"

    cell = crystal.cell
    space_group = crystal.space_group
    atom_site = crystal.atom_site
    space_group_symop = space_group.full_space_group_symop
    atom_site_susceptibility = crystal.atom_site_susceptibility
    # l_magnetic_labes = atom_site_susceptibility.label

    volume = density_point.volume_unit_cell
    np = density_point.number_unit_cell

    den_i = density_point.numpy_density
    den_ferro_i = density_point.numpy_density_ferro
    den_antiferro_i = density_point.numpy_density_antiferro
    mult_i = density_point.numpy_multiplicity

    l_f_nucl, l_fr_e, l_fr_s, l_e_up, l_h_loc = [], [], [], [], []
    l_k_hkl, l_phase_3d = [], []
    l_chi_perp_ferro, l_chi_perp_antiferro = [], []
    total_peaks = 0
    for diffrn in l_diffrn:
        diffrn_orient_matrix = diffrn.diffrn_orient_matrix
        u_matrix = diffrn_orient_matrix.u
        e_up = calc_e_up_loc(0., 0., 0., u_matrix)

        setup = diffrn.setup
        field = float(setup.field)
        h_loc = (field*e_up[0], field*e_up[1], field*e_up[2])
        diffrn_refln = diffrn.diffrn_refln
        index_h = numpy.array(diffrn_refln.index_h, dtype=int)
        index_k = numpy.array(diffrn_refln.index_k, dtype=int)
        index_l = numpy.array(diffrn_refln.index_l, dtype=int)
        total_peaks += index_h.size
        hkl = (index_h, index_k, index_l)
        fr_e = numpy.array(diffrn_refln.fr, dtype=float)
        fr_s = numpy.array(diffrn_refln.fr_sigma, dtype=float)
        index_hkl = numpy.stack([index_h, index_k, index_l], dtype=bool)
        f_nucl = crystal.calc_f_nucl(index_hkl)
        k_hkl = cell.calc_k_loc(*hkl)
        phase_3d = density_point.calc_phase_3d(hkl, space_group_symop)

        moment_2d, chi_2d_ferro, chi_2d_antiferro = \
            density_point.calc_moment_2d(
                space_group_symop, cell, atom_site_susceptibility, h_loc,
                chi_iso_ferro=1., chi_iso_antiferro=1.,
                flag_two_channel=flag_two_channel)

        chi_ferro = calc_fm_by_density(mult_i, den_ferro_i, np, volume,
                                       chi_2d_ferro, phase_3d)
        chi_perp_ferro = calc_moment_perp(k_hkl, chi_ferro)

        chi_aferro = calc_fm_by_density(mult_i, den_antiferro_i, np, volume,
                                        chi_2d_antiferro, phase_3d)
        chi_perp_aferro = calc_moment_perp(k_hkl, chi_aferro)

        l_f_nucl.append(f_nucl)
        l_fr_e.append(fr_e)
        l_fr_s.append(fr_s)
        l_e_up.append(e_up)
        l_h_loc.append(h_loc)
        l_k_hkl.append(k_hkl)
        l_phase_3d.append(phase_3d)
        l_chi_perp_ferro.append(chi_perp_ferro)
        l_chi_perp_antiferro.append(chi_perp_aferro)

    l_name_1 = atom_site_susceptibility.get_variable_names()
    l_par_1_0 = [atom_site_susceptibility.get_variable_by_name(name) for name
                 in l_name_1]

    l_name_2 = mem_parameters.get_variable_names()
    l_par_2_0 = [mem_parameters.get_variable_by_name(name) for name
                 in l_name_2]
    l_name = l_name_1 + l_name_2
    l_par_0 = l_par_1_0 + l_par_2_0

    if len(l_name) == 0:
        return

    def temp_func(l_par):
        for name, parameter in zip(l_name, l_par):
            if name[0][0] == "atom_site_susceptibility":
                atom_site_susceptibility.set_variable_by_name(name, parameter)
            elif name[0][0] == "mem_parameters":
                mem_parameters.set_variable_by_name(name, parameter)

        chi_iso_f = mem_parameters.chi_ferro
        chi_iso_af = mem_parameters.chi_antiferro

        atom_site_susceptibility.apply_chi_iso_constraint(cell)
        atom_site_susceptibility.apply_space_group_constraint(
            atom_site, space_group, cell)

        l_chi_sq = []
        l_der_chi_sq, l_der_chi_sq_f, l_der_chi_sq_a = [], [], []

        # FIXME: add flag for orbital magnetic moment
        for diffrn, f_nucl, fr_e, fr_s, e_up, h_loc, k_hkl, phase_3d, \
            chi_perp_ferro, chi_perp_aferro in \
            zip(l_diffrn, l_f_nucl, l_fr_e, l_fr_s, l_e_up, l_h_loc, l_k_hkl,
                l_phase_3d, l_chi_perp_ferro, l_chi_perp_antiferro):

            moment_2d, moment_ferro, moment_antiferro = \
                density_point.calc_moment_2d(
                    space_group_symop, cell, atom_site_susceptibility, h_loc,
                    chi_iso_ferro=1., chi_iso_antiferro=1.,
                    flag_two_channel=flag_two_channel)

            f_m = calc_fm_by_density(mult_i, den_i, np, volume, moment_2d,
                                     phase_3d)
            f_m_perp = calc_moment_perp(k_hkl, f_m)

            # # correction on orbital moment
            # diffrn_refln = diffrn.diffrn_refln
            # setup = diffrn.setup
            # field = setup.field
            # ind_h = diffrn_refln.numpy_index_h
            # ind_k = diffrn_refln.numpy_index_k
            # ind_l = diffrn_refln.numpy_index_l
            # chi_m = crystal.calc_susceptibility_moment_tensor(
            #     ind_h, ind_k, ind_l, flag_only_orbital=True)
            # sft_ij = chi_m[:9]
            # sftm_ij = chi_m[9:]

            # fm_orb_perp_loc = calc_fm_perp_loc(e_up, field, k_hkl, sft_ij,
            #                                    sftm_ij)

            # add ferro and anti_ferro

            f_m_perp_sum = (
                f_m_perp[0] + chi_iso_f*chi_perp_ferro[0] +
                chi_iso_af*chi_perp_aferro[0],
                f_m_perp[1] + chi_iso_f*chi_perp_ferro[1] +
                chi_iso_af*chi_perp_aferro[1],
                f_m_perp[2] + chi_iso_f*chi_perp_ferro[2] +
                chi_iso_af*chi_perp_aferro[2])

            fr_m, delta_fr_m = diffrn.calc_fr(cell, f_nucl, f_m_perp_sum,
                                              delta_f_m_perp=f_m_perp)
            delta_fr_m_f = delta_fr_m
            delta_fr_m_a = delta_fr_m

            diffrn.diffrn_refln.numpy_fr_calc = fr_m

            chi_sq, der_chi_sq = calc_chi_sq(fr_e, fr_s, fr_m, delta_fr_m)
            der_chi_sq_f = calc_chi_sq(fr_e, fr_s, fr_m, delta_fr_m_f)[1]
            der_chi_sq_a = calc_chi_sq(fr_e, fr_s, fr_m, delta_fr_m_a)[1]
            l_chi_sq.append(chi_sq)
            l_der_chi_sq.append(der_chi_sq)
            l_der_chi_sq_f.append(der_chi_sq_f)
            l_der_chi_sq_a.append(der_chi_sq_a)
        return sum(l_chi_sq)

    chi_sq = temp_func(l_par_0)
    print(f"Chi_sq before optimization {chi_sq/total_peaks:.5f}.           ",
          end="\r")
    if flag_info:
        d_info["print"] = \
            f"Chi_sq/n before optimization {chi_sq/total_peaks:.5f}."

    res = scipy.optimize.minimize(
        temp_func, l_par_0, method="BFGS",
        callback=lambda x: func_temp(x, param_name=l_name, d_info=d_info),
        options={"eps": 0.001})

    l_param = res.x
    chi_sq_new = res.fun

    hess_inv = res["hess_inv"]
    sigma = (abs(numpy.diag(hess_inv)))**0.5

    print(f"Chi_sq after optimization {chi_sq_new/total_peaks:.5f}.          ",
          end="\r")
    if flag_info:
        d_info["print"] = \
            f"Chi_sq/n after optimization {chi_sq_new/total_peaks:.5f}."

    for name, parameter, sig in zip(l_name, l_param, sigma):
        name_sig = name[:-1] + ((f"{name[-1][0]:}_sigma", name[-1][1]), )
        if name[0][0] == "atom_site_susceptibility":
            atom_site_susceptibility.set_variable_by_name(name, parameter)
            atom_site_susceptibility.set_variable_by_name(name_sig, sig)
        elif name[0][0] == "mem_parameters":
            mem_parameters.set_variable_by_name(name, parameter)
            mem_parameters.set_variable_by_name(name_sig, sig)

    for diffrn in l_diffrn:
        diffrn.diffrn_refln.numpy_to_items()
        chi_sq, points = diffrn.diffrn_refln.calc_chi_sq_points()
        refine_ls = RefineLs(goodness_of_fit_all=chi_sq/points,
                             number_reflns=points)
        diffrn.add_items([refine_ls])


def func_temp(*argv, param_name: List[tuple] = None, d_info: dict = None) \
        -> bool:
    """Show data at optimization procedure."""
    flag_out = False
    l_param = argv[0]
    s_out = " ".join([f"{_:10.5f}" for _ in l_param])
    print(f"{s_out:}", end="\r")
    if d_info is not None:
        if param_name is not None:
            ls_out = [f"{name[-1][0]:}: {param: 10.5f}" for name, param in
                      zip(param_name, l_param)]
        else:
            ls_out = [f"{param: 10.5f}" for param in l_param]
        d_info["print"] = "Best solution:\n\n"+"\n".join(ls_out)
        flag_out = d_info["stop"]
    return flag_out


def make_cycle(crystal: Crystal, l_diffrn: List[Diffrn],
               mem_parameters: MEMParameters,
               c_lambda: float = 1e-6, n_iterations: int = 51,
               n_cycle: int = 10, disp: bool = True,
               d_info: dict = None):
    """Rho - Chi cycle."""
    flag_info = d_info is not None

    d_info_2 = None
    if flag_info:
        d_info_keys = d_info.keys()
        if "stop" not in d_info_keys:
            d_info["stop"] = False
        if "print" not in d_info_keys:
            d_info["print"] = ""
        if "d_info" not in d_info_keys:
            d_info_2 = {"stop": False, "print": ""}
            d_info["d_info"] = d_info_2

    flag_info = d_info is not None

    for i_cycle in range(n_cycle):
        print(f"Cycle {(i_cycle+1):4}. Entropy maximization                  ",
              end="\r")

        if flag_info:
            d_info["print"] = f"""Cycle {(i_cycle+1):4}/{n_cycle:}:
    Entropy maximization."""
            if d_info["stop"]:
                d_info["stop"] = False
                break
        density_point = maximize_entropy(
            crystal, l_diffrn, mem_parameters, c_lambda=c_lambda,
            n_iterations=n_iterations, disp=disp, d_info=d_info_2)

        if flag_info:
            d_info["print"] = f"""Cycle {(i_cycle+1):4}/{n_cycle:}:
    Chi refinement."""
            if d_info["stop"]:
                d_info["stop"] = False
                break
        print(f"Cycle {(i_cycle+1):4}. Chi refinement                        ",
              end="\r")
        refine_susceptibility(crystal, l_diffrn, density_point, mem_parameters,
                              disp=disp, d_info=d_info_2)
    return density_point


def calc_moments(fract_x, fract_y, fract_z, field_loc,
                 density_point: DensityPointL, crystal: Crystal,
                 mem_parameters: MEMParameters):
    """Calculate magnetic moments in points fract_x, fract_y, fract_z."""
    points_a = mem_parameters.points_a
    points_b = mem_parameters.points_b
    points_c = mem_parameters.points_c

    cell = crystal.cell
    atom_site = crystal.atom_site

    np_ind_xyz = numpy.transpose(numpy.array([fract_x*points_a, fract_y*points_b, fract_z*points_c], dtype=float))

    # np_ind_xyz = (fract_x*points_a, fract_y*points_b, fract_z*points_c)

    chi_3d_11, chi_3d_22, chi_3d_33, chi_3d_12, chi_3d_13, chi_3d_23, \
        chi_3d_iso = calc_densities_3d_for_mem(
        density_point, crystal, mem_parameters=mem_parameters)

    c_11 = tri_linear_interpolation(np_ind_xyz, chi_3d_11)
    c_22 = tri_linear_interpolation(np_ind_xyz, chi_3d_22)
    c_33 = tri_linear_interpolation(np_ind_xyz, chi_3d_33)
    c_12 = tri_linear_interpolation(np_ind_xyz, chi_3d_12)
    c_13 = tri_linear_interpolation(np_ind_xyz, chi_3d_13)
    c_23 = tri_linear_interpolation(np_ind_xyz, chi_3d_23)

    c_iso = tri_linear_interpolation(np_ind_xyz, chi_3d_iso)

    c_orto = cell.ortogonalize_matrix((c_11, c_12, c_13, c_12, c_22, c_23,
                                       c_13, c_23, c_33))
    m_x, m_y, m_z = calc_product_matrix_vector(c_orto, field_loc)
    m_b_x, m_b_y, m_b_z = c_iso*field_loc[0], c_iso*field_loc[1], \
        c_iso*field_loc[2]

    moment_x, moment_y, moment_z = m_x+m_b_x, m_y+m_b_y, m_z+m_b_z

    return moment_x, moment_y, moment_z


def calc_moments_in_unit_cell(density_point: DensityPointL, crystal: Crystal,
                              mem_parameters: MEMParameters, field_loc):
    """Calculate magnetization density in an unit cell."""
    points_a = mem_parameters.points_a
    points_b = mem_parameters.points_b
    points_c = mem_parameters.points_c

    h_x = numpy.linspace(0., 1., num=points_a, endpoint=False, dtype=float)
    h_y = numpy.linspace(0., 1., num=points_b, endpoint=False, dtype=float)
    h_z = numpy.linspace(0., 1., num=points_c, endpoint=False, dtype=float)

    fract_x, fract_y, fract_z = numpy.meshgrid(h_x, h_y, h_z, indexing="ij")
    fract_x, fract_y = fract_x.flatten(), fract_y.flatten()
    fract_z = fract_z.flatten()

    moment_x, moment_y, moment_z = calc_moments(
        fract_x, fract_y, fract_z, field_loc, density_point, crystal,
        mem_parameters)

    return fract_x, fract_y, fract_z, moment_x, moment_y, moment_z


def calc_section_for_mem(section: Section, density_point: DensityPointL,
                         crystal: Crystal, mem_parameters: MEMParameters):
    """
    Calculate magnetization density of paramagnetic compound at given field.

    The result is written into file.
    """
    field_loc = numpy.array([section.field_x, section.field_y,
                             section.field_z], dtype=float)
    cell = crystal.cell
    atom_site = crystal.atom_site
    np_fract_x, np_fract_y, np_fract_z = section.calc_fractions(
        cell, atom_site)

    moment_x, moment_y, moment_z = calc_moments(
        np_fract_x, np_fract_y, np_fract_z, field_loc, density_point, crystal,
        mem_parameters)

    size_x, size_y = float(section.size_x), float(section.size_y)

    np_x = (numpy.array(range(section.points_x), dtype=float) -
            0.5*section.points_x) * size_x / float(section.points_x)

    np_y = (numpy.array(range(section.points_y), dtype=float) -
            0.5*section.points_y) * size_y / float(section.points_y)

    np_x_2d, np_y_2d = numpy.meshgrid(np_x, np_y, indexing="ij")
    np_x, np_y = np_x_2d.flatten(), np_y_2d.flatten()

    v_pos_x, v_pos_y, v_pos_z = section.calc_axes_x_y_z(cell, atom_site)

    ls_out = ["loop_\n_section_cut_dist_x\n_section_cut_dist_y"]
    ls_out.append("_section_cut_moment_x\n_section_cut_moment_y")
    ls_out.append("_section_cut_moment_z")
    for _x, _y, v_m_x, v_m_y, v_m_z in zip(
            np_x, np_y, moment_x, moment_y, moment_z):
        val_x = v_pos_x[0]*v_m_x + v_pos_x[1]*v_m_y + v_pos_x[2]*v_m_z
        val_y = v_pos_y[0]*v_m_x + v_pos_y[1]*v_m_y + v_pos_y[2]*v_m_z
        val_z = v_pos_z[0]*v_m_x + v_pos_z[1]*v_m_y + v_pos_z[2]*v_m_z
        ls_out.append(f"{_x:9.5f} {_y:9.5f} {val_x:15.10f} {val_y:15.10f} \
{val_z:15.10f}")
    with open(section.url_out, "w") as fid:
        fid.write("\n".join(ls_out))


def calc_densities_3d_for_mem(density_point: DensityPointL, crystal: Crystal,
                              mem_parameters: MEMParameters):
    """
    Calculate 3d density for susceptibility tensor and isotropical background.

    Output:

        - chi_den_3d_11, chi_den_3d_22, chi_den_3d_33,
          chi_den_3d_12, chi_den_3d_13, chi_den_3d_23,
          chi_den_3d_iso
    """
    points_a = mem_parameters.points_a
    points_b = mem_parameters.points_b
    points_c = mem_parameters.points_c
    chi_iso_ferro = mem_parameters.chi_ferro
    chi_iso_antiferro = mem_parameters.chi_antiferro

    atom_site_susceptibility = crystal.atom_site_susceptibility

    space_group = crystal.space_group
    full_space_group_symop = space_group.full_space_group_symop

    points_abc = (points_a, points_b, points_c)

    density_point.calc_rbs_i(full_space_group_symop, points_a=points_a,
                             points_b=points_b, points_c=points_c)

    susc_11, susc_22, susc_33, susc_12, susc_13, susc_23 = \
        density_point.calc_susc_i(atom_site_susceptibility)

    index_xyz = (numpy.array(density_point.index_x, dtype=int),
                 numpy.array(density_point.index_y, dtype=int),
                 numpy.array(density_point.index_z, dtype=int))

    den_i = numpy.array(density_point.density, dtype=float)
    den_ferro_i = numpy.array(density_point.density_ferro, dtype=float)
    den_aferro_i = numpy.array(density_point.density_antiferro, dtype=float)

    r_11 = full_space_group_symop.numpy_r_11.astype(float)
    r_12 = full_space_group_symop.numpy_r_12.astype(float)
    r_13 = full_space_group_symop.numpy_r_13.astype(float)
    r_21 = full_space_group_symop.numpy_r_21.astype(float)
    r_22 = full_space_group_symop.numpy_r_22.astype(float)
    r_23 = full_space_group_symop.numpy_r_23.astype(float)
    r_31 = full_space_group_symop.numpy_r_31.astype(float)
    r_32 = full_space_group_symop.numpy_r_32.astype(float)
    r_33 = full_space_group_symop.numpy_r_33.astype(float)

    b_1 = full_space_group_symop.numpy_b_1.astype(float)
    b_2 = full_space_group_symop.numpy_b_2.astype(float)
    b_3 = full_space_group_symop.numpy_b_3.astype(float)

    r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
    b_i = (b_1, b_2, b_3)

    den_3d = transfer_to_density_3d(index_xyz, den_i, points_abc, r_ij, b_i)
    den_ferro_3d = transfer_to_density_3d(
        index_xyz, den_ferro_i, points_abc, r_ij, b_i)
    den_aferro_3d = transfer_to_density_3d(
        index_xyz, den_aferro_i, points_abc, r_ij, b_i)

    chi_3d_11, chi_3d_22, chi_3d_33, chi_3d_12, chi_3d_13, chi_3d_23 = \
        transfer_to_chi_3d(index_xyz, susc_11, susc_22, susc_33, susc_12,
                           susc_13, susc_23, points_abc, r_ij, b_i)

    chi_den_3d_11, chi_den_3d_22 = chi_3d_11*den_3d, chi_3d_22*den_3d
    chi_den_3d_33, chi_den_3d_12 = chi_3d_33*den_3d, chi_3d_12*den_3d
    chi_den_3d_13, chi_den_3d_23 = chi_3d_13*den_3d, chi_3d_23*den_3d

    chi_den_3d_iso = chi_iso_ferro*den_ferro_3d + \
        chi_iso_antiferro*den_aferro_3d

    return chi_den_3d_11, chi_den_3d_22, chi_den_3d_33,\
        chi_den_3d_12, chi_den_3d_13, chi_den_3d_23, chi_den_3d_iso



"""Described classes AtomTypeScat, AtomTypeScatL."""
import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_1_magnetic import \
    get_j0_j2_by_symbol
from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

class AtomTypeScat(ItemN):
    """
    AtomTypeScat class.

    Data items in the ATOM_TYPE_SCAT category describe atomic
    scattering information used in crystallographic structure studies.
    This category is fully defined in the core CIF dictionary.

    Attributes
    ----------
        - symbol (mandatory)
        - neutron_magnetic_j0_a1, neutron_magnetic_j0_a2
        - neutron_magnetic_j0_b1, neutron_magnetic_j0_b2
        - neutron_magnetic_j0_c1, neutron_magnetic_j0_c2
        - neutron_magnetic_j0_d
        - neutron_magnetic_j2_a1, neutron_magnetic_j2_a2
        - neutron_magnetic_j2_b1, neutron_magnetic_j2_b2
        - neutron_magnetic_j2_c1, neutron_magnetic_j2_c2
        - neutron_magnetic_j2_d
        - neutron_magnetic_j4_a1, neutron_magnetic_j4_a2
        - neutron_magnetic_j4_b1, neutron_magnetic_j4_b2
        - neutron_magnetic_j4_c1, neutron_magnetic_j4_c2
        - neutron_magnetic_j4_d
        - neutron_magnetic_j6_a1, neutron_magnetic_j6_a2
        - neutron_magnetic_j6_b1, neutron_magnetic_j6_b2
        - neutron_magnetic_j6_c1, neutron_magnetic_j6_c2
        - neutron_magnetic_j6_d

    Methods
    -------
        - calc_form_factor
        - calc_j0
        - calc_j2
        - load_by_symbol
        - form_by_symbol (classmethod)
    """

    ATTR_MANDATORY_NAMES = ("symbol", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("symbol", )

    ATTR_OPTIONAL_NAMES = ("neutron_magnetic_j0_a1", "neutron_magnetic_j0_a2",
                           "neutron_magnetic_j0_b1", "neutron_magnetic_j0_b2",
                           "neutron_magnetic_j0_c1", "neutron_magnetic_j0_c2",
                           "neutron_magnetic_j0_d",  "neutron_magnetic_j0_e",
                           "neutron_magnetic_j2_a1", "neutron_magnetic_j2_a2",
                           "neutron_magnetic_j2_b1", "neutron_magnetic_j2_b2",
                           "neutron_magnetic_j2_c1", "neutron_magnetic_j2_c2",
                           "neutron_magnetic_j2_d",  "neutron_magnetic_j2_e",
                           "neutron_magnetic_j4_a1", "neutron_magnetic_j4_a2",
                           "neutron_magnetic_j4_b1", "neutron_magnetic_j4_b2",
                           "neutron_magnetic_j4_c1", "neutron_magnetic_j4_c2",
                           "neutron_magnetic_j4_d",  "neutron_magnetic_j4_e",
                           "neutron_magnetic_j6_a1", "neutron_magnetic_j6_a2",
                           "neutron_magnetic_j6_b1", "neutron_magnetic_j6_b2",
                           "neutron_magnetic_j6_c1", "neutron_magnetic_j6_c2",
                           "neutron_magnetic_j6_d",  "neutron_magnetic_j6_e",
                           "neutron_magnetic_source")
    ATTR_OPTIONAL_TYPES = (
        float, float, float, float, float, float, float, float, float, float,
        float, float, float, float, float, float, float, float, float, float,
        float, float, float, float, float, float, float, float, float, float,
        float, float, str)
    ATTR_OPTIONAL_CIF = ("neutron_magnetic_j0_a1", "neutron_magnetic_j0_a2",
                         "neutron_magnetic_j0_b1", "neutron_magnetic_j0_b2",
                         "neutron_magnetic_j0_c1", "neutron_magnetic_j0_c2",
                         "neutron_magnetic_j0_d",  "neutron_magnetic_j0_e",
                         "neutron_magnetic_j2_a1", "neutron_magnetic_j2_a2",
                         "neutron_magnetic_j2_b1", "neutron_magnetic_j2_b2",
                         "neutron_magnetic_j2_c1", "neutron_magnetic_j2_c2",
                         "neutron_magnetic_j2_d",  "neutron_magnetic_j2_e",
                         "neutron_magnetic_j4_a1", "neutron_magnetic_j4_a2",
                         "neutron_magnetic_j4_b1", "neutron_magnetic_j4_b2",
                         "neutron_magnetic_j4_c1", "neutron_magnetic_j4_c2",
                         "neutron_magnetic_j4_d",  "neutron_magnetic_j4_e",
                         "neutron_magnetic_j6_a1", "neutron_magnetic_j6_a2",
                         "neutron_magnetic_j6_b1", "neutron_magnetic_j6_b2",
                         "neutron_magnetic_j6_c1", "neutron_magnetic_j6_c2",
                         "neutron_magnetic_j6_d",  "neutron_magnetic_j6_e",
                         "neutron_magnetic_source")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("j0_parameters", "j2_parameters")
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_type_scat"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomTypeScat, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_form_factor(self, sthovl, lande=2., kappa=1.,
                         flag_only_orbital=False):
        r"""Calculate magnetic form factor in frame of Spherical model.

        (Int.Tabl.C.p.592)

        :LFactor:  is Lande factor
        :lsthovl:  is list :math:`sin(\\theta)/\\lambda` in :math:`\\A^{-1}`

        Calculation of magnetic form factor <j0>, <j2>, <j4>, <j6>

        :coeff: is a list [A, a, B, b, C, c, D] at n = 0, 2, 4, 6

        `<https://journals.aps.org/prb/pdf/10.1103/PhysRevB.79.140405>`
        mismatch with international tables where (1.0-2.0/np_factor_lande)
        """
        # not sure about kappa, it is here just for test, by default it is 1.0
        j2_av = self.calc_j2(sthovl, kappa=kappa)

        if flag_only_orbital:
            form_factor = (2.0/float(lande)-1.0)*j2_av
        else:
            j0_av = self.calc_j0(sthovl, kappa=kappa)
            form_factor = j0_av+(2.0/float(lande)-1.0)*j2_av
        return form_factor

    def calc_j0(self, sthovl, kappa: float = 1):
        """Calculate j0."""
        j0_A = self.neutron_magnetic_j0_a1
        j0_a = self.neutron_magnetic_j0_a2
        j0_B = self.neutron_magnetic_j0_b1
        j0_b = self.neutron_magnetic_j0_b2
        j0_C = self.neutron_magnetic_j0_c1
        j0_c = self.neutron_magnetic_j0_c2
        j0_D = self.neutron_magnetic_j0_d
        _h = (sthovl/float(kappa))**2
        j0_av = (j0_A*numpy.exp(-j0_a*_h) +
                 j0_B*numpy.exp(-j0_b*_h) +
                 j0_C*numpy.exp(-j0_c*_h)+j0_D)
        return j0_av

    def calc_j2(self, sthovl, kappa: float = 1):
        """Calculate j2."""
        j2_A = self.neutron_magnetic_j2_a1
        j2_a = self.neutron_magnetic_j2_a2
        j2_B = self.neutron_magnetic_j2_b1
        j2_b = self.neutron_magnetic_j2_b2
        j2_C = self.neutron_magnetic_j2_c1
        j2_c = self.neutron_magnetic_j2_c2
        j2_D = self.neutron_magnetic_j2_d
        _h = (sthovl/float(kappa))**2
        j2_av = (j2_A*numpy.exp(-j2_a*_h) +
                 j2_B*numpy.exp(-j2_b*_h) +
                 j2_C*numpy.exp(-j2_c*_h) + j2_D)*_h
        return j2_av

    def load_by_symbol(self, symbol: str = None) -> NoReturn:
        """Load by symbol."""
        if symbol is None:
            symbol = self.symbol
        else:
            self.symbol = symbol
        j0_A1, j0_a2, j0_B1, j0_b2, j0_C1, j0_c2, j0_D, j2_A1, j2_a2, j2_B1, \
            j2_b2, j2_C1, j2_c2, j2_D = get_j0_j2_by_symbol(symbol)
        self.neutron_magnetic_j0_a1 = j0_A1
        self.neutron_magnetic_j0_a2 = j0_a2
        self.neutron_magnetic_j0_b1 = j0_B1
        self.neutron_magnetic_j0_b2 = j0_b2
        self.neutron_magnetic_j0_c1 = j0_C1
        self.neutron_magnetic_j0_c2 = j0_c2
        self.neutron_magnetic_j0_d = j0_D
        self.neutron_magnetic_j2_a1 = j2_A1
        self.neutron_magnetic_j2_a2 = j2_a2
        self.neutron_magnetic_j2_b1 = j2_B1
        self.neutron_magnetic_j2_b2 = j2_b2
        self.neutron_magnetic_j2_c1 = j2_C1
        self.neutron_magnetic_j2_c2 = j2_c2
        self.neutron_magnetic_j2_d = j2_D

        self.j0_parameters = numpy.array([j0_A1, j0_a2, j0_B1, j0_b2, j0_C1, j0_c2, j0_D], dtype=float)
        self.j2_parameters = numpy.array([j2_A1, j2_a2, j2_B1, j2_b2, j2_C1, j2_c2, j2_D], dtype=float)

    @classmethod
    def form_by_symbol(cls, symbol: str):
        """Form by symbol (classmethod)."""
        j0_A1, j0_a2, j0_B1, j0_b2, j0_C1, j0_c2, j0_D, j2_A1, j2_a2, j2_B1,\
            j2_b2, j2_C1, j2_c2, j2_D = get_j0_j2_by_symbol(symbol)
        item = cls(
            symbol=symbol,
            neutron_magnetic_j0_a1=j0_A1, neutron_magnetic_j0_a2=j0_a2,
            neutron_magnetic_j0_b1=j0_B1, neutron_magnetic_j0_b2=j0_b2,
            neutron_magnetic_j0_c1=j0_C1, neutron_magnetic_j0_c2=j0_c2,
            neutron_magnetic_j0_d=j0_D,
            neutron_magnetic_j2_a1=j2_A1, neutron_magnetic_j2_a2=j2_a2,
            neutron_magnetic_j2_b1=j2_B1, neutron_magnetic_j2_b2=j2_b2,
            neutron_magnetic_j2_c1=j2_C1, neutron_magnetic_j2_c2=j2_c2,
            neutron_magnetic_j2_d=j2_D)
        item.j0_parameters = numpy.array([j0_A1, j0_a2, j0_B1, j0_b2, j0_C1, j0_c2, j0_D], dtype=float)
        item.j2_parameters = numpy.array([j2_A1, j2_a2, j2_B1, j2_b2, j2_C1, j2_c2, j2_D], dtype=float)
        return item


class AtomTypeScatL(LoopN):
    """Description of AtomTypeScat in loop."""

    ITEM_CLASS = AtomTypeScat
    ATTR_INDEX = "symbol"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomTypeScatL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def calc_form_factor(self, sthovl, l_lande: list, l_kappa: list,
                         flag_only_orbital: bool = False):
        """Calculate form-factor."""
        form_factor = [item.calc_form_factor(
            sthovl, float(lande), float(kappa),
            flag_only_orbital=flag_only_orbital) for item, lande, kappa in
            zip(self.items, l_lande, l_kappa)]
        return numpy.array(form_factor, dtype=float)

    @classmethod
    def form_by_symbols(cls, symbols: list):
        """Form by symbols (classmethod)."""
        items = []
        item_class = cls.ITEM_CLASS
        for symbol in symbols:
            items.append(item_class.form_by_symbol(symbol))
        obj = cls()
        obj.items = items
        return obj

# obj = AtomTypeScatL.form_by_symbols(["O2-", "Fe3+"])
# print(obj, end="\n\n")
